import numpy as np
import pickle
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from matplotlib.cm import ScalarMappable
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg, NavigationToolbar2QT
from mpl_toolkits.basemap import Basemap
from mayavi.core.ui.api import MayaviScene, MlabSceneModel, SceneEditor
from traits.api import HasTraits, Instance
from traitsui.api import View, Item
from astropy import units as au, constants as ac
from PyQt5 import QtCore, QtWidgets


groupBoxStyle = '''
QGroupBox {
    font-weight: bold;
    background-color: white;
    border: 1px solid green;
    border-radius: 2px;
    margin-top: 20px;
           }

QGroupBox::title {
    subcontrol-origin: margin;
    subcontrol-position: top center;
    padding-left: 10px;
    padding-right: 10px;
    padding-top: 5px;
                  }
'''


class MplCanvas(FigureCanvasQTAgg):

    def __init__(self, tight_layout=False, projection=None, kind=None, figsize=None):
        self.fig = Figure(tight_layout=tight_layout, figsize=figsize)

        if kind == '1':
            ax1 = plt.subplot2grid((3, 1), (0, 0), rowspan=2)
            ax2 = plt.subplot2grid((3, 1), (2, 0))

            self.fig.add_subplot(ax1.get_subplotspec())
            self.fig.add_subplot(ax2.get_subplotspec())

        elif kind == '2':

            gs_kw = dict(width_ratios=[1, 1], height_ratios=[0.5, 20])
            _, axd = plt.subplot_mosaic([['upper center', 'upper center'],
                                         ['left', 'right']], gridspec_kw=gs_kw, layout="constrained")

            self.fig.add_subplot(axd['left'].get_subplotspec())
            self.fig.add_subplot(axd['right'].get_subplotspec())
            self.fig.add_subplot(axd['upper center'].get_subplotspec())

        else:
            self.axes = self.fig.add_subplot(111, projection=projection)

        super(MplCanvas, self).__init__(self.fig)


class BaseCanvas(FigureCanvasQTAgg):
    def __init__(self, incl):
        self.fig = Figure()

        gs_kw = dict(width_ratios=[1, 1], height_ratios=[1, 1])
        _, axd = plt.subplot_mosaic([['upper left', 'upper right'],
                                     ['lower left', 'lower right']], gridspec_kw=gs_kw,
                                    layout="constrained")

        self.fig.add_subplot(axd['upper left'].get_subplotspec())
        self.fig.add_subplot(axd['upper right'].get_subplotspec())
        self.fig.add_subplot(axd['lower left'].get_subplotspec())
        self.fig.add_subplot(axd['lower right'].get_subplotspec())

        super(BaseCanvas, self).__init__(self.fig)

        self.m1 = Basemap(ax=self.fig.axes[0], projection='ortho', lat_0=90 - incl, lon_0=0.0)
        self.m2 = Basemap(ax=self.fig.axes[1], projection='ortho', lat_0=90 - incl, lon_0=-90)
        self.m3 = Basemap(ax=self.fig.axes[2], projection='ortho', lat_0=90 - incl, lon_0=-180)
        self.m4 = Basemap(ax=self.fig.axes[3], projection='ortho', lat_0=90 - incl, lon_0=-270)


class Visualization(HasTraits):
    scene = Instance(MlabSceneModel, ())

    view = View(Item('scene', editor=SceneEditor(scene_class=MayaviScene),
                     show_label=False), resizable=True)


class PlotGUI(QtWidgets.QMainWindow):

    def __init__(self, DIP, plot_params):

        super().__init__()

        self.DIP = DIP

        self.ints = DIP.opt_results['ints']
        self.mapprojs = DIP.mapprojs

        self.vmin = (DIP.params['Tcool'] / DIP.params['Tphot']) ** 4
        self.vmax = (DIP.params['Thot'] / DIP.params['Tphot']) ** 4

        line_chisq = DIP.opt_stats['Chi-square for Line Profile(s)']
        mol1_chisq = DIP.opt_stats['Chi-square for Molecular(1) Profile(s)']
        mol2_chisq = DIP.opt_stats['Chi-square for Molecular(2) Profile(s)']
        lc_chisq = DIP.opt_stats['Chi-square for Light Curve Profile']
        alpha_line_chisq = DIP.opt_stats['Alpha * Line Profile(s) Chi-square']
        beta_mol1_chisq = DIP.opt_stats['Beta * Molecular(1) Profile(s) Chi-square']
        gamma_mol2_chisq = DIP.opt_stats['Gamma * Molecular(2) Profile(s) Chi-square']
        delta_lc_chisq = DIP.opt_stats['Delta * Light Curve Profile Chi-square']
        total_chisq = DIP.opt_stats['Total Weighted Chi-square']
        mem = DIP.opt_stats['Total Entropy']
        lmbd_mem = DIP.opt_stats['Lambda * Total Entropy']
        ftot = DIP.opt_stats['Loss Function Value']

        lmbd = DIP.opt_results['lmbd']
        alpha = DIP.opt_results['alpha']
        beta = DIP.opt_results['beta']
        gamma = DIP.opt_results['gamma']
        delta = DIP.opt_results['delta']
        nit = int(DIP.opt_results['nit'])
        nfev = int(DIP.opt_results['nfev'])

        csa = DIP.opt_results['total_cool_spotted_area']
        hsa = DIP.opt_results['total_hot_spotted_area']
        psa = DIP.opt_results['total_unspotted_area']

        self.surface_grid = DIP.surface_grid.copy()

        self.line_sp = plot_params['line_sep_prf']
        self.line_sr = plot_params['line_sep_res']
        self.mol_sp = plot_params['mol_sep_prf']
        self.mol_sr = plot_params['mol_sep_res']
        self.seb = plot_params['show_err_bars']

        self.ms = plot_params['markersize']
        self.fs = plot_params['fontsize']
        self.lw = plot_params['linewidth']
        self.tls = plot_params['ticklabelsize']

        self.setWindowTitle('SpotDIPy - Plot and Result Window')
        self.resize(1024, 768)

        centralWidget = QtWidgets.QWidget(self)
        centralWidgetVLayout = QtWidgets.QVBoxLayout(centralWidget)
        self.setCentralWidget(centralWidget)

        saveOMDataButton = QtWidgets.QPushButton('Save Obs. and Model Data')
        saveMaPDataButton = QtWidgets.QPushButton('Save Map Projections Data')
        saveGUIButton = QtWidgets.QPushButton('Save GUI')
        saveButtonsHLayout = QtWidgets.QHBoxLayout()
        saveButtonsHLayout.addWidget(saveOMDataButton)
        saveButtonsHLayout.addWidget(saveMaPDataButton)
        saveButtonsHLayout.addWidget(saveGUIButton)

        main_tab = QtWidgets.QTabWidget(self)

        centralWidgetVLayout.addWidget(main_tab)
        centralWidgetVLayout.addLayout(saveButtonsHLayout)

        profile_tab = QtWidgets.QTabWidget(self)

        line_profile_frame = QtWidgets.QFrame()
        line_profile_frame_glayout = QtWidgets.QGridLayout(line_profile_frame)

        line_profile_gbox = QtWidgets.QGroupBox('Line Profiles')
        line_profile_gbox_glayout = QtWidgets.QGridLayout(line_profile_gbox)
        line_profile_gbox_glayout.setContentsMargins(4, 4, 4, 4)
        line_profile_gbox_glayout.setSpacing(4)
        line_profile_gbox.setStyleSheet(groupBoxStyle)

        line_sep_profiles_label = QtWidgets.QLabel('Profile Seperation :')
        line_sep_redisuals_label = QtWidgets.QLabel('Residual Seperation :')

        self.line_sep_profiles_ledit = QtWidgets.QLineEdit(str(self.line_sp))
        self.line_sep_residuals_ledit = QtWidgets.QLineEdit(str(self.line_sr))

        self.line_error_bar_cbox = QtWidgets.QCheckBox('Show Errorbar')
        self.line_error_bar_cbox.setLayoutDirection(QtCore.Qt.RightToLeft)
        self.line_error_bar_cbox.setChecked(True if self.seb else False)

        line_replot_profiles_button = QtWidgets.QPushButton('Replot')

        line_profile_settings_hlayout = QtWidgets.QHBoxLayout()
        line_profile_settings_hlayout.addStretch()
        line_profile_settings_hlayout.addWidget(line_sep_profiles_label)
        line_profile_settings_hlayout.addWidget(self.line_sep_profiles_ledit)
        line_profile_settings_hlayout.addWidget(line_sep_redisuals_label)
        line_profile_settings_hlayout.addWidget(self.line_sep_residuals_ledit)
        line_profile_settings_hlayout.addWidget(self.line_error_bar_cbox)
        line_profile_settings_hlayout.addWidget(line_replot_profiles_button)
        line_profile_settings_hlayout.addStretch()

        line_profile_frame_glayout.addWidget(line_profile_gbox, 0, 0, 1, 1)
        line_profile_frame_glayout.addLayout(line_profile_settings_hlayout, 1, 0, 1, 2)

        self.line_profile_plot = MplCanvas(tight_layout=True, kind='2', figsize=None)
        line_profile_toolbar = NavigationToolbar2QT(self.line_profile_plot, self)

        line_profile_gbox_glayout.addWidget(line_profile_toolbar)
        line_profile_gbox_glayout.addWidget(self.line_profile_plot)

        """"""" lc """""""
        lc_profile_frame = QtWidgets.QFrame()
        lc_profile_frame_glayout = QtWidgets.QGridLayout(lc_profile_frame)

        lc_profile_gbox = QtWidgets.QGroupBox('Light Curve Profile')
        lc_profile_gbox_glayout = QtWidgets.QGridLayout(lc_profile_gbox)
        lc_profile_gbox_glayout.setContentsMargins(4, 4, 4, 4)
        lc_profile_gbox_glayout.setSpacing(4)
        lc_profile_gbox.setStyleSheet(groupBoxStyle)

        lc_residual_gbox = QtWidgets.QGroupBox('Residuals')
        lc_residual_gbox_glayout = QtWidgets.QGridLayout(lc_residual_gbox)
        lc_residual_gbox_glayout.setContentsMargins(4, 4, 4, 4)
        lc_residual_gbox_glayout.setSpacing(4)
        lc_residual_gbox.setStyleSheet(groupBoxStyle)

        self.lc_error_bar_cbox = QtWidgets.QCheckBox('Show Errorbar')
        self.lc_error_bar_cbox.setLayoutDirection(QtCore.Qt.RightToLeft)
        self.lc_error_bar_cbox.setChecked(True if self.seb else False)

        lc_profile_settings_hlayout = QtWidgets.QHBoxLayout()
        lc_profile_settings_hlayout.addStretch()
        lc_profile_settings_hlayout.addWidget(self.lc_error_bar_cbox)
        lc_profile_settings_hlayout.addStretch()

        self.lc_profile_plot = MplCanvas(tight_layout=True, kind='1')
        lc_profile_toolbar = NavigationToolbar2QT(self.lc_profile_plot, self)

        lc_profile_gbox_glayout.addWidget(lc_profile_toolbar)
        lc_profile_gbox_glayout.addWidget(self.lc_profile_plot)

        lc_profile_frame_glayout.addWidget(lc_profile_gbox, 0, 0,)
        lc_profile_frame_glayout.addLayout(lc_profile_settings_hlayout, 1, 0)

        """"""" mol 1 """""""

        mol1_profile_frame = QtWidgets.QFrame()
        mol1_profile_frame_glayout = QtWidgets.QGridLayout(mol1_profile_frame)

        mol1_profile_gbox = QtWidgets.QGroupBox('Molecular Profiles - 1')
        mol1_profile_gbox_glayout = QtWidgets.QGridLayout(mol1_profile_gbox)
        mol1_profile_gbox_glayout.setContentsMargins(4, 4, 4, 4)
        mol1_profile_gbox_glayout.setSpacing(4)
        mol1_profile_gbox.setStyleSheet(groupBoxStyle)

        mol1_sep_profiles_label = QtWidgets.QLabel('Profile Seperation :')
        mol1_sep_redisuals_label = QtWidgets.QLabel('Residual Seperation :')

        self.mol1_sep_profiles_ledit = QtWidgets.QLineEdit(str(self.mol_sp))
        self.mol1_sep_residuals_ledit = QtWidgets.QLineEdit(str(self.mol_sr))

        self.mol1_error_bar_cbox = QtWidgets.QCheckBox('Show Errorbar')
        self.mol1_error_bar_cbox.setLayoutDirection(QtCore.Qt.RightToLeft)
        self.mol1_error_bar_cbox.setChecked(True if self.seb else False)

        mol1_replot_profiles_button = QtWidgets.QPushButton('Replot')

        mol1_profile_settings_hlayout = QtWidgets.QHBoxLayout()
        mol1_profile_settings_hlayout.addStretch()
        mol1_profile_settings_hlayout.addWidget(mol1_sep_profiles_label)
        mol1_profile_settings_hlayout.addWidget(self.mol1_sep_profiles_ledit)
        mol1_profile_settings_hlayout.addWidget(mol1_sep_redisuals_label)
        mol1_profile_settings_hlayout.addWidget(self.mol1_sep_residuals_ledit)
        mol1_profile_settings_hlayout.addWidget(self.mol1_error_bar_cbox)
        mol1_profile_settings_hlayout.addWidget(mol1_replot_profiles_button)
        mol1_profile_settings_hlayout.addStretch()

        mol1_profile_frame_glayout.addWidget(mol1_profile_gbox, 0, 0, 1, 1)
        mol1_profile_frame_glayout.addLayout(mol1_profile_settings_hlayout, 1, 0, 1, 2)

        self.mol1_profile_plot = MplCanvas(tight_layout=True, kind='2', figsize=(2, 1))
        mol1_profile_toolbar = NavigationToolbar2QT(self.mol1_profile_plot, self)

        mol1_profile_gbox_glayout.addWidget(mol1_profile_toolbar)
        mol1_profile_gbox_glayout.addWidget(self.mol1_profile_plot)

        """"""

        """"""" mol 2 """""""

        mol2_profile_frame = QtWidgets.QFrame()
        mol2_profile_frame_glayout = QtWidgets.QGridLayout(mol2_profile_frame)

        mol2_profile_gbox = QtWidgets.QGroupBox('Molecular Profiles - 2')
        mol2_profile_gbox_glayout = QtWidgets.QGridLayout(mol2_profile_gbox)
        mol2_profile_gbox_glayout.setContentsMargins(4, 4, 4, 4)
        mol2_profile_gbox_glayout.setSpacing(4)
        mol2_profile_gbox.setStyleSheet(groupBoxStyle)

        mol2_sep_profiles_label = QtWidgets.QLabel('Profile Seperation :')
        mol2_sep_redisuals_label = QtWidgets.QLabel('Residual Seperation :')

        self.mol2_sep_profiles_ledit = QtWidgets.QLineEdit(str(self.mol_sp))
        self.mol2_sep_residuals_ledit = QtWidgets.QLineEdit(str(self.mol_sr))

        self.mol2_error_bar_cbox = QtWidgets.QCheckBox('Show Errorbar')
        self.mol2_error_bar_cbox.setLayoutDirection(QtCore.Qt.RightToLeft)
        self.mol2_error_bar_cbox.setChecked(True if self.seb else False)

        mol2_replot_profiles_button = QtWidgets.QPushButton('Replot')

        mol2_profile_settings_hlayout = QtWidgets.QHBoxLayout()
        mol2_profile_settings_hlayout.addStretch()
        mol2_profile_settings_hlayout.addWidget(mol2_sep_profiles_label)
        mol2_profile_settings_hlayout.addWidget(self.mol2_sep_profiles_ledit)
        mol2_profile_settings_hlayout.addWidget(mol2_sep_redisuals_label)
        mol2_profile_settings_hlayout.addWidget(self.mol2_sep_residuals_ledit)
        mol2_profile_settings_hlayout.addWidget(self.mol2_error_bar_cbox)
        mol2_profile_settings_hlayout.addWidget(mol2_replot_profiles_button)
        mol2_profile_settings_hlayout.addStretch()

        mol2_profile_frame_glayout.addWidget(mol2_profile_gbox, 0, 0, 1, 1)
        mol2_profile_frame_glayout.addLayout(mol2_profile_settings_hlayout, 1, 0, 1, 2)

        self.mol2_profile_plot = MplCanvas(tight_layout=True, kind='2', figsize=(2, 1))
        mol2_profile_toolbar = NavigationToolbar2QT(self.mol2_profile_plot, self)

        mol2_profile_gbox_glayout.addWidget(mol2_profile_toolbar)
        mol2_profile_gbox_glayout.addWidget(self.mol2_profile_plot)

        """"""

        profile_tab.addTab(line_profile_frame, 'Line Profiles')
        profile_tab.addTab(mol1_profile_frame, 'Molecular Profiles - 1')
        profile_tab.addTab(mol2_profile_frame, 'Molecular Profiles - 2')
        profile_tab.addTab(lc_profile_frame, 'Light Curve Profile')
        profile_tab.setTabVisible(0, False)
        profile_tab.setTabVisible(1, False)
        profile_tab.setTabVisible(2, False)
        profile_tab.setTabVisible(3, True)

        self.mercator_gbox = QtWidgets.QGroupBox('Mercator Projection')
        mercator_gbox_glayout = QtWidgets.QGridLayout(self.mercator_gbox)
        mercator_gbox_glayout.setContentsMargins(0, 0, 0, 0)
        mercator_gbox_glayout.setSpacing(0)
        self.mercator_gbox.setStyleSheet(groupBoxStyle)

        self.mercator_plot = MplCanvas()
        mercator_toolbar = NavigationToolbar2QT(self.mercator_plot, self)

        mercator_gbox_glayout.addWidget(mercator_toolbar)
        mercator_gbox_glayout.addWidget(self.mercator_plot)

        self.mollweide_gbox = QtWidgets.QGroupBox('Mollweide Projection')
        mollweide_gbox_glayout = QtWidgets.QGridLayout(self.mollweide_gbox)
        mollweide_gbox_glayout.setContentsMargins(0, 0, 0, 0)
        mollweide_gbox_glayout.setSpacing(0)
        self.mollweide_gbox.setStyleSheet(groupBoxStyle)

        self.mollweide_plot = MplCanvas(projection="mollweide")
        mollweide_toolbar = NavigationToolbar2QT(self.mollweide_plot, self)

        mollweide_gbox_glayout.addWidget(mollweide_toolbar)
        mollweide_gbox_glayout.addWidget(self.mollweide_plot)

        self.spherical_gbox = QtWidgets.QGroupBox('Spherical Projection')
        spherical_gbox_glayout = QtWidgets.QGridLayout(self.spherical_gbox)
        spherical_gbox_glayout.setContentsMargins(0, 0, 0, 0)
        spherical_gbox_glayout.setSpacing(0)
        self.spherical_gbox.setStyleSheet(groupBoxStyle)

        self.spherical_plot = BaseCanvas(incl=DIP.params['incl'])
        spherical_toolbar = NavigationToolbar2QT(self.spherical_plot, self)

        spherical_gbox_glayout.addWidget(spherical_toolbar)
        spherical_gbox_glayout.addWidget(self.spherical_plot)

        self.trid_plot_gbox = QtWidgets.QGroupBox('3D Surface Map')
        trid_plot_gbox_glayout = QtWidgets.QGridLayout(self.trid_plot_gbox)
        trid_plot_gbox_glayout.setContentsMargins(4, 4, 4, 4)
        trid_plot_gbox_glayout.setSpacing(4)
        self.trid_plot_gbox.setStyleSheet(groupBoxStyle)

        self.trid_plot_vis = Visualization()
        self.trid_plot = self.trid_plot_vis.edit_traits(parent=self, kind='subpanel').control

        trid_dis_label = QtWidgets.QLabel('Cam. Distance :')
        trid_incl_label = QtWidgets.QLabel('Inclination :')
        trid_phase_label = QtWidgets.QLabel('Phase :')

        self.trid_incl_dsbox = QtWidgets.QDoubleSpinBox()
        self.trid_incl_dsbox.setMinimum(0)
        self.trid_incl_dsbox.setMaximum(90)
        self.trid_incl_dsbox.setSingleStep(1)
        self.trid_phase_dsbox = QtWidgets.QDoubleSpinBox()
        self.trid_phase_dsbox.setMinimum(0)
        self.trid_phase_dsbox.setMaximum(1)
        self.trid_phase_dsbox.setDecimals(4)
        self.trid_phase_dsbox.setSingleStep(0.01)
        self.trid_dis_dsbox = QtWidgets.QDoubleSpinBox()
        self.trid_dis_dsbox.setValue(8)
        self.trid_dis_dsbox.setMinimum(0)
        self.trid_dis_dsbox.setSingleStep(1.0)

        trid_plot_gbox_glayout.addWidget(self.trid_plot, 0, 0, 1, 6)
        trid_plot_gbox_glayout.addWidget(trid_dis_label, 1, 0, 1, 1)
        trid_plot_gbox_glayout.addWidget(self.trid_dis_dsbox, 1, 1, 1, 1)
        trid_plot_gbox_glayout.addWidget(trid_incl_label, 1, 2, 1, 1)
        trid_plot_gbox_glayout.addWidget(self.trid_incl_dsbox, 1, 3, 1, 1)
        trid_plot_gbox_glayout.addWidget(trid_phase_label, 1, 4, 1, 1)
        trid_plot_gbox_glayout.addWidget(self.trid_phase_dsbox, 1, 5, 1, 1)

        maps_tab = QtWidgets.QTabWidget(self)

        maps_tab.addTab(self.mollweide_gbox, 'Mollweide Projection')
        maps_tab.addTab(self.mercator_gbox, 'Mercator Projection')
        maps_tab.addTab(self.spherical_gbox, 'Spherical Projection')
        maps_tab.addTab(self.trid_plot_gbox, '3D Projection')

        results_tab = QtWidgets.QTabWidget(self)

        result_frame = QtWidgets.QFrame()
        result_frame_glayout = QtWidgets.QGridLayout(result_frame)

        opt_result_gbox = QtWidgets.QGroupBox('Optimization Result')
        opt_result_gbox.setStyleSheet(groupBoxStyle)
        opt_result_gbox_glayout = QtWidgets.QGridLayout(opt_result_gbox)

        opt_results_row_names = ['Chi-square for Line Profile(s)', 'Chi-square for Molecular(1) Profile(s)',
                                 'Chi-square for Molecular(2) Profile(s)', 'Chi-square for Light Curve Profile',
                                 'alpha * Line Profile(s) Chi-square', 'beta * Molecular(1) Profile(s) Chi-square',
                                 'gamma * Molecular(2) Profile(s) Chi-square', 'delta * Light Curve Profile Chi-square',
                                 'Total Chi-square', 'Total Entropy', 'lmbd * Total Entropy',
                                 'Loss Function Value', 'Number of iteration(s)', 'Number of function evaluation(s)',
                                 'alpha', 'beta', 'gamma', 'delta', 'lmbd']
        opt_results_row_vals = [line_chisq, mol1_chisq, mol2_chisq, lc_chisq, alpha_line_chisq, beta_mol1_chisq,
                                gamma_mol2_chisq, delta_lc_chisq, total_chisq, mem, lmbd_mem, ftot, nit, nfev, alpha,
                                beta, gamma, delta, lmbd]

        opt_result_table = QtWidgets.QTableWidget(len(opt_results_row_names), 1)
        opt_result_table.setEditTriggers(QtWidgets.QTableWidget.NoEditTriggers)
        opt_result_table.horizontalHeader().setVisible(False)
        for rni, (row_name, row_val) in enumerate(zip(opt_results_row_names, opt_results_row_vals)):
            opt_result_table.setVerticalHeaderItem(rni, QtWidgets.QTableWidgetItem(row_name))
            row_val = row_val if type(row_val) is int else round(row_val, 6)
            opt_result_table.setItem(rni, 0, QtWidgets.QTableWidgetItem(str(row_val)))
        opt_result_table.horizontalHeader().setSectionResizeMode(QtWidgets.QHeaderView.Stretch)
        opt_result_table.resizeColumnsToContents()

        opt_result_gbox_glayout.addWidget(opt_result_table)

        map_information_row_names = ['Number of Surface Elements', 'Surface Discritization Method',
                                     'Total Spotted Area (%)', 'Total Cool Spotted Area (%)',
                                     'Total Hot Spotted Area (%)', 'Total Unspotted Area (%)',
                                     'Minimum Reduced Intensity (I/Iphot)', 'Maximum Reduced Intensity (I/Iphot)']
        map_information_row_vals = [DIP.surface_grid['noes'], DIP.surface_grid['method'], round((csa + hsa) * 100, 6),
                                    round(csa * 100, 6), round(hsa * 100, 6), round(psa * 100, 6), round(self.vmin, 6),
                                    round(self.vmax, 6)]
        map_information_gbox = QtWidgets.QGroupBox('Stellar Surface Information')
        map_information_gbox.setStyleSheet(groupBoxStyle)
        map_information_gbox_glayout = QtWidgets.QGridLayout(map_information_gbox)

        map_information_table = QtWidgets.QTableWidget(len(map_information_row_names), 1)
        map_information_table.setEditTriggers(QtWidgets.QTableWidget.NoEditTriggers)
        map_information_table.horizontalHeader().setVisible(False)
        for rni, (row_name, row_val) in enumerate(zip(map_information_row_names, map_information_row_vals)):
            map_information_table.setVerticalHeaderItem(rni, QtWidgets.QTableWidgetItem(row_name))
            map_information_table.setItem(rni, 0, QtWidgets.QTableWidgetItem(str(row_val)))

        map_information_table.horizontalHeader().setSectionResizeMode(QtWidgets.QHeaderView.Stretch)
        map_information_table.resizeColumnsToContents()

        map_information_gbox_glayout.addWidget(map_information_table)

        lamda_search_gbox = QtWidgets.QGroupBox('Lambda Search Plot')
        lamda_search_gbox.setStyleSheet(groupBoxStyle)
        lamda_search_gbox_glayout = QtWidgets.QGridLayout(lamda_search_gbox)

        self.lamda_search_plot = MplCanvas(tight_layout=True)
        lamda_search_toolbar = NavigationToolbar2QT(self.lamda_search_plot, self)

        lamda_search_gbox_glayout.addWidget(lamda_search_toolbar)
        lamda_search_gbox_glayout.addWidget(self.lamda_search_plot)

        result_frame_glayout.addWidget(opt_result_gbox, 0, 0, 1, 1)
        result_frame_glayout.addWidget(map_information_gbox, 0, 1, 1, 1)

        results_tab.addTab(result_frame, 'Optimization Results')
        results_tab.addTab(lamda_search_gbox, 'Lambda Search')
        results_tab.setTabVisible(1, False)

        main_tab.addTab(profile_tab, 'Observations, Models and Residuals')
        main_tab.addTab(maps_tab, 'Surface Map Projections')
        main_tab.addTab(results_tab, 'Results')

        self.trid_phase_dsbox.valueChanged.connect(self.set_3d_plot_ori)
        self.trid_incl_dsbox.valueChanged.connect(self.set_3d_plot_ori)
        self.trid_dis_dsbox.valueChanged.connect(self.set_3d_plot_ori)

        line_replot_profiles_button.clicked.connect(self.plot_line_profiles)
        mol1_replot_profiles_button.clicked.connect(self.plot_mol1_profiles)
        mol2_replot_profiles_button.clicked.connect(self.plot_mol2_profiles)
        self.line_error_bar_cbox.stateChanged.connect(self.plot_line_profiles)
        self.lc_error_bar_cbox.stateChanged.connect(self.plot_lc_profile)
        self.mol1_error_bar_cbox.stateChanged.connect(self.plot_mol1_profiles)
        self.mol2_error_bar_cbox.stateChanged.connect(self.plot_mol2_profiles)

        saveOMDataButton.clicked.connect(self.save_obs_model_data)
        saveMaPDataButton.clicked.connect(self.save_map_data)
        saveGUIButton.clicked.connect(self.saveGUI)

        if 'mol2' in DIP.conf:
            profile_tab.setTabVisible(2, True)
            self.plot_mol2_profiles()
            profile_tab.setCurrentIndex(2)

        if 'mol1' in DIP.conf:
            profile_tab.setTabVisible(1, True)
            self.plot_mol1_profiles()
            profile_tab.setCurrentIndex(1)

        if 'line' in DIP.conf:
            profile_tab.setTabVisible(0, True)
            self.plot_line_profiles()
            profile_tab.setCurrentIndex(0)

        self.plot_lc_profile()
        self.plot_mercator(crmap=self.mapprojs['rmap'], cextent=np.rad2deg(self.mapprojs['extent']))
        self.plot_mollweide(crmap=self.mapprojs['rmap'], xlats=self.mapprojs['xlats'], xlongs=self.mapprojs['xlongs'])
        self.plot_spherical(crmap=self.mapprojs['rmap'], mlats=self.mapprojs['mlats'], mlongs=self.mapprojs['mlongs'])
        self.plot_3d_surface()

        if 'lmbds' in DIP.opt_results and len(DIP.opt_results['lmbds']) > 1:
            results_tab.setTabVisible(1, True)

            lmbds = DIP.opt_results['lmbds']
            chisqs = DIP.opt_results['total_wchisqs']
            mems = DIP.opt_results['mems']
            maxcurve = DIP.opt_results['maxcurve']
            self.plot_lambda_search(lmbds=lmbds, chisqs=chisqs, mems=mems, maxcurve=maxcurve)

    def save_obs_model_data(self):

        path, _ = QtWidgets.QFileDialog.getSaveFileName()

        if path:
            try:
                self.DIP.save_obs_model_data(path)
                QtWidgets.QMessageBox.information(self, 'SpotDIPy', 'The data has been saved!')
            except IOError:
                QtWidgets.QMessageBox.information(self, 'SpotDIPy', 'The data has NOT been saved!')

    def save_map_data(self):

        path, _ = QtWidgets.QFileDialog.getSaveFileName()

        if path:
            try:
                self.DIP.save_map_data(path)
                QtWidgets.QMessageBox.information(self, 'SpotDIPy', 'The data has been saved!')
            except IOError:
                QtWidgets.QMessageBox.information(self, 'SpotDIPy', 'The data has NOT been saved!')

    def saveGUI(self):

        path, _ = QtWidgets.QFileDialog.getSaveFileName()

        if path:
            try:
                file = open(path, 'wb')
                pickle.dump(self.DIP, file)
                file.close()

                QtWidgets.QMessageBox.information(self, 'SpotDIPy', 'The GUI has been saved!')
            except IOError:
                QtWidgets.QMessageBox.information(self, 'SpotDIPy', 'The GUI has NOT been saved!')

    def plot_line_profiles(self):

        spl_line_slps = self.DIP.opt_results['line']['spotless_sprfs']
        rcs_line_slps = self.DIP.opt_results['line']['recons_sprfs']

        line_sp = float(self.line_sep_profiles_ledit.text())
        line_sr = float(self.line_sep_residuals_ledit.text())
        seb = self.line_error_bar_cbox.checkState()

        fig = self.line_profile_plot.fig

        ax1 = fig.axes[0]
        ax2 = fig.axes[1]
        ax3 = fig.axes[2]

        ax1.cla()
        ax2.cla()
        ax3.cla()

        ax3.axis("off")

        for i, time in enumerate(spl_line_slps):

            epoch = (time - self.DIP.params['t0']) / self.DIP.params['period']

            obs_line_prf = self.DIP.idc['line']['data'][time]['prf']
            obs_line_err = self.DIP.idc['line']['data'][time]['errs']
            spl_line_prf = spl_line_slps[time]['prf']
            rcs_line_prf = rcs_line_slps[time]['prf']

            maxv = max(self.DIP.idc['line']['vels'])
            maxi = max(obs_line_prf) + i * line_sp
            residual = obs_line_prf - rcs_line_prf
            maxir = np.average(residual + i * line_sr)
            xy1 = (maxv - maxv / 3.1, maxi + line_sp / 10.)
            xy2 = (maxv - maxv / 10., maxir + line_sr / 10.)

            if seb:
                ax1.errorbar(self.DIP.idc['line']['vels'], obs_line_prf + i * line_sp, yerr=obs_line_err, fmt='o',
                             color='k', ms=self.ms, zorder=1)
                ax2.errorbar(self.DIP.idc['line']['vels'], residual + i * line_sr, yerr=obs_line_err, fmt='o',
                             color='k', ms=self.ms, zorder=1)
            else:
                ax1.plot(self.DIP.idc['line']['vels'], obs_line_prf + i * line_sp, 'ko', ms=self.ms, zorder=1)
                ax2.plot(self.DIP.idc['line']['vels'], residual + i * line_sr, 'ko', ms=self.ms, zorder=1)

            ax1.plot(self.DIP.idc['line']['vels'], spl_line_prf + i * line_sp, 'b', linewidth=self.lw, zorder=2)
            ax1.plot(self.DIP.idc['line']['vels'], rcs_line_prf + i * line_sp, 'r', linewidth=self.lw, zorder=3)
            ax1.annotate(str('%0.3f' % round(epoch, 3)), xy=xy1, color='g')
            ax2.annotate(str('%0.3f' % round(epoch, 3)), xy=xy2, color='g')
            ax2.axhline(i * line_sr, color='r', zorder=2)

        ax1.axvline(0.0, color='g')

        ax3.plot([], [], 'ko', label='Observed Line Profiles', ms=self.ms)
        ax3.plot([], [], 'b', label='Spotless Model', linewidth=self.lw)
        ax3.plot([], [], 'r', label='Spotted Model', linewidth=self.lw)
        ax3.legend(loc='upper center', ncol=3, bbox_to_anchor=(0.5, 1.02), frameon=False)

        ax1.set_xlabel('Velocity (km/s)', fontsize=self.fs)
        ax1.set_ylabel('$\mathregular{I/I_c}$', fontsize=self.fs)

        ax2.set_xlabel('Velocity (km/s)', fontsize=self.fs)
        ax2.set_ylabel('Residuals', fontsize=self.fs)

        ax1.tick_params(axis='both', labelsize=self.tls)
        ax2.tick_params(axis='both', labelsize=self.tls)

        fig.subplots_adjust(hspace=0)

        self.line_profile_plot.draw()

    def plot_lc_profile(self):

        seb = self.lc_error_bar_cbox.checkState()

        ax1 = self.lc_profile_plot.fig.axes[0]
        ax1.cla()

        ax2 = self.lc_profile_plot.fig.axes[1]
        ax2.cla()

        recons_slc = self.DIP.opt_results['lc']['recons_slc']

        if 'lc' in self.DIP.conf:

            times = self.DIP.idc['lc']['times']
            fluxs = self.DIP.idc['lc']['data']['fluxs']
            errs = self.DIP.idc['lc']['data']['errs']

            epochs = (times - self.DIP.params['t0']) / self.DIP.params['period']

            if seb:
                ax1.errorbar(epochs, fluxs, yerr=errs, fmt='o', color='k', label='Observed Light Curve', ms=self.ms,
                             zorder=1)
                ax2.errorbar(epochs, fluxs - recons_slc, yerr=errs, fmt='o', color='k', ms=self.ms, zorder=1)
            else:
                ax1.plot(epochs, fluxs, 'ko', label='Observed Light Curve', ms=self.ms, zorder=1)
                ax2.plot(epochs, fluxs - recons_slc, 'ko', ms=self.ms, zorder=1)

            ax1.plot(epochs, recons_slc, 'r', label='Spotted Model', linewidth=self.lw, zorder=2)

        else:
            ntimes = self.DIP.opt_results['lc']['ntimes']
            nepochs = (ntimes - self.DIP.params['t0']) / self.DIP.params['period']

            ax1.plot(nepochs, recons_slc, 'ko', ms=2, label='Synthetic Light Curve')

        ax1.set_xlabel('Epoch', fontsize=self.fs)
        ax1.set_ylabel('Normalized Flux', fontsize=self.fs)

        ax2.set_xlabel('Epoch', fontsize=self.fs)
        ax2.set_ylabel('Residuals', fontsize=self.fs)

        ax1.tick_params(axis='both', labelsize=self.tls)
        ax2.tick_params(axis='both', labelsize=self.tls)

        ax2.axhline(0.0, color='r', linewidth=self.lw, zorder=2)

        ax1.legend(loc='upper center', bbox_to_anchor=(0.5, 1.10), ncol=2, frameon=False)

        self.lc_profile_plot.draw()

    def plot_mol1_profiles(self):

        self.plot_mol_profiles(self.mol1_sep_profiles_ledit,
                               self.mol1_sep_residuals_ledit,
                               self.mol1_error_bar_cbox,
                               self.mol1_profile_plot,
                               self.DIP.idc['mol1']['data'],
                               self.DIP.idc['mol1']['vels'],
                               np.average(self.DIP.conf['mol1']['wave_range']),
                               'mol1')

    def plot_mol2_profiles(self):

        self.plot_mol_profiles(self.mol2_sep_profiles_ledit,
                               self.mol2_sep_residuals_ledit,
                               self.mol2_error_bar_cbox,
                               self.mol2_profile_plot,
                               self.DIP.idc['mol2']['data'],
                               self.DIP.idc['mol2']['vels'],
                               np.average(self.DIP.conf['mol2']['wave_range']),
                               'mol2')

    def plot_mol_profiles(self, sep_prf_ledit, sep_res_ledit, err_cbox, prf_plot, mol_obs_data,
                          mol_vels, mwave, mode):

        rcs_mol_slps = self.DIP.opt_results[mode]['recons_sprfs']

        mol_sp = float(sep_prf_ledit.text())
        mol_sr = float(sep_res_ledit.text())
        seb = err_cbox.checkState()

        ax1 = prf_plot.fig.axes[0]
        ax2 = prf_plot.fig.axes[1]
        ax3 = prf_plot.fig.axes[2]

        ax1.cla()
        ax2.cla()
        ax3.cla()

        ax3.axis("off")

        mol_waves = (mol_vels * mwave) / ac.c.to(au.kilometer / au.second).value + mwave

        for i, time in enumerate(rcs_mol_slps):

            epoch = (time - self.DIP.params['t0']) / self.DIP.params['period']

            obs_mol_prf = mol_obs_data[time]['prf']
            obs_mol_err = mol_obs_data[time]['errs']
            rcs_mol_prf = rcs_mol_slps[time]['prf']

            maxv = max(mol_waves)
            minv = min(mol_waves)
            maxi = np.average(obs_mol_prf[int(len(mol_waves)*0.9):]) + i * mol_sp
            residual = obs_mol_prf - rcs_mol_prf
            maxir = np.average(residual + i * mol_sr)
            # xy1 = (maxv - maxv / 3.1, maxi + mol_sp / 10.)
            xy1 = (maxv + (maxv - minv) * 0.01, maxi)
            xy2 = (maxv - (maxv - minv) * 0.01, maxir + maxir / 10.)

            if seb:
                ax1.errorbar(mol_waves, obs_mol_prf + i * mol_sp, yerr=obs_mol_err, fmt='o', color='k', ms=self.ms)
                ax2.errorbar(mol_waves, residual + i * mol_sr, yerr=obs_mol_err, fmt='o', color='k', ms=self.ms)
            else:
                ax1.plot(mol_waves, obs_mol_prf + i * mol_sp, 'k', ms=self.ms)
                ax2.plot(mol_waves, residual + i * mol_sr, 'k', ms=self.ms)

            ax1.plot(mol_waves, rcs_mol_prf + i * mol_sp, 'r', linewidth=self.lw, zorder=3)
            ax1.annotate(str('%0.3f' % round(epoch, 3)), xy=xy1, color='g')
            ax2.annotate(str('%0.3f' % round(epoch, 3)), xy=xy2, color='g')
            ax2.axhline(i * mol_sr, color='r', zorder=3)

        ax3.plot([], [], 'ko', label='Observed Molecular Profile', ms=self.ms)
        ax3.plot([], [], 'r', label='Spotted Model', linewidth=self.lw)
        ax3.legend(loc='upper center', ncol=3, bbox_to_anchor=(0.5, 1.02), frameon=False)

        ax1.set_xlabel('Wavelength ($\AA$)', fontsize=self.fs)
        ax1.set_ylabel('Normalized Flux', fontsize=self.fs)
        # ax1.legend(loc='upper center', bbox_to_anchor=(0.5, 1.10), ncol=3, frameon=False)

        ax2.set_xlabel('Wavelength ($\AA$)', fontsize=self.fs)
        ax2.set_ylabel('Residuals', fontsize=self.fs)

        ax1.tick_params(axis='both', labelsize=self.tls)
        ax2.tick_params(axis='both', labelsize=self.tls)

        prf_plot.draw()

    def plot_mercator(self, crmap, cextent):

        fig = self.mercator_plot.fig
        ax = fig.axes[0]

        img = ax.imshow(crmap, cmap='gist_heat', aspect='equal', extent=cextent, interpolation='bicubic',
                        vmin=self.vmin, vmax=self.vmax)

        ax.fill_between(x=[cextent[0], cextent[1]], y1=-self.DIP.params['incl'], y2=cextent[2], color='k', alpha=0.3)

        ax.set_xticks(np.arange(0, 420, 60))
        ax.set_yticks(np.arange(-90, 120, 30))
        ax.set_xlabel('Longitude ($^\circ$)', fontsize=self.fs)
        ax.set_ylabel('Latitude ($^\circ$)', fontsize=self.fs)

        ax.grid(True)

        colors = ['b', 'white', 'g', 'purple']
        for i, mode in enumerate(self.DIP.conf):
            epochs = (self.DIP.idc[mode]['times'] - self.DIP.params['t0']) / self.DIP.params['period']
            phases = epochs - np.floor(epochs)
            obslong = 360 * (1.0 - phases)

            ax.plot([obslong, obslong], [-85, -75], '-', color=colors[i], linewidth=1)
            if 0.0 in phases:
                ax.plot([0, 0], [-85, -75], '-', color=colors[i], linewidth=1)

        ax.tick_params(axis='both', labelsize=self.tls)

        cb = fig.colorbar(img, ax=ax, location='bottom', shrink=0.5)
        cb.set_label(r'$\frac{I}{I_{phot}}$', fontsize=self.fs + 5)
        cb.ax.tick_params(labelsize=self.tls)
        cb.set_ticks(np.linspace(self.vmin, self.vmax, 6))

        fig.tight_layout()

        self.mercator_plot.draw()

    def plot_mollweide(self, crmap, xlongs, xlats):

        fig = self.mollweide_plot.fig
        ax = fig.axes[0]

        img = ax.pcolormesh(xlongs - np.pi, xlats, crmap, cmap='gist_heat', shading='gouraud',
                            vmin=self.vmin, vmax=self.vmax)

        ax.fill_between(x=[-np.pi, np.pi], y1=np.deg2rad(-self.DIP.params['incl']), y2=-np.pi / 2., color='k',
                        alpha=0.3)

        ax.set_xticks(np.deg2rad(np.arange(-120, 180, 60)))
        ax.set_yticks(np.deg2rad(np.arange(-90, 120, 30)))
        xtick_labels = np.arange(60, 360, 60)
        ax.set_xticklabels(xtick_labels, zorder=15)
        ax.grid(True)

        ax.tick_params(labelsize=self.tls)
        ax.xaxis.set_label_coords(0.5, -0.100)

        colors = ['b', 'white', 'g', 'purple']
        for i, mode in enumerate(self.DIP.conf):
            epochs = (self.DIP.idc[mode]['times'] - self.DIP.params['t0']) / self.DIP.params['period']
            phases = epochs - np.floor(epochs)
            obslong = 360 * (1.0 - phases) - 180

            ax.plot(np.deg2rad([obslong, obslong]), np.deg2rad([-30, -20]), '-', color=colors[i], linewidth=1)
            if 0.0 in phases:
                ax.plot(np.deg2rad([0.0, 0.0]), np.deg2rad([-30, -20]), '-', color=colors[i], linewidth=1)

        ax.set_xlabel('Longitude ($^\circ$)', fontsize=self.fs)
        ax.set_ylabel('Latitude ($^\circ$)', fontsize=self.fs)

        cb = self.mollweide_plot.fig.colorbar(img, ax=ax, location='bottom', shrink=0.5)
        cb.set_label(r'$\frac{I}{I_{phot}}$', fontsize=self.fs + 5)
        cb.ax.tick_params(labelsize=self.tls)
        cb.set_ticks(np.linspace(self.vmin, self.vmax, 6))

        fig.tight_layout()

        self.mollweide_plot.draw()

    def plot_spherical(self, crmap, mlats, mlongs):

        mlats = np.rad2deg(mlats)
        mlongs = np.rad2deg(mlongs)

        minp, maxp = np.min(mlats), np.max(mlats)
        minm, maxm = np.min(mlongs), np.max(mlongs)

        mr = np.linspace(minm, maxm, 11)
        pr = np.linspace(minp, maxp, 7)

        fig = self.spherical_plot.fig
        axs = self.spherical_plot.fig.axes

        ms = [self.spherical_plot.m1, self.spherical_plot.m2,
              self.spherical_plot.m3, self.spherical_plot.m4]
        for i, (m, ax) in enumerate(zip(ms, axs)):

            m.drawmeridians(mr, linewidth=1.0)
            m.drawparallels(pr, linewidth=1.0)

            x, y = m(mlongs, mlats)
            img = m.contourf(x, y, crmap, 100, cmap='gist_heat', vmin=self.vmin, vmax=self.vmax)
            ax.set_title('Phase = ' + str('%0.2f' % (i * 0.25)), pad=10)

        fig.tight_layout()

        cb = fig.colorbar(ScalarMappable(norm=img.norm, cmap=img.cmap), ax=self.spherical_plot.fig.axes,
                          ticks=range(0, 0 + 5, 5), location='right', shrink=0.75)
        cb.set_label(r'$\frac{I}{I_{phot}}$', fontsize=self.fs + 5)
        cb.ax.tick_params(labelsize=self.tls)
        cb.set_ticks(np.linspace(self.vmin, self.vmax, 6))

        self.spherical_plot.draw()

    def plot_3d_surface(self):

        fss = self.ints.copy()
        grid_xyzs = self.surface_grid['grid_xyzs'].copy()
        if self.DIP.surface_grid['method'] != 'phoebe2_marching':
            fss = np.repeat(self.ints, 2)

        tdx = []
        tdy = []
        tdz = []
        triangles = []
        scalars = []
        for i, xyz in enumerate(grid_xyzs):
            scalars.append([fss[i]] * 3)
            for row in xyz:
                tdx.append(row[0])
                tdy.append(row[1])
                tdz.append(row[2])
            triangles.append((0 + i * 3, 1 + i * 3, 2 + i * 3))
        tdmap = np.hstack(scalars)

        tmesh = self.trid_plot_vis.scene.mlab.triangular_mesh(tdx, tdy, tdz, triangles,  scalars=tdmap,
                                                              colormap='gist_heat', line_width=3.0, vmin=self.vmin,
                                                              vmax=self.vmax)

        tmesh.scene.y_plus_view()

        cb = self.trid_plot_vis.scene.mlab.colorbar(tmesh, nb_labels=6, label_fmt='%0.1f', orientation='vertical')
        cb.label_text_property.font_family = 'times'
        cb.label_text_property.bold = 0
        cb.label_text_property.font_size = 2

        self.trid_incl_dsbox.setValue(self.DIP.params['incl'])
        self.set_3d_plot_ori()

    def set_3d_plot_ori(self):

        azimuth = -1 * (self.trid_phase_dsbox.value() * 360)
        elevation = self.trid_incl_dsbox.value()
        distance = self.trid_dis_dsbox.value()

        self.trid_plot_vis.scene.mlab.view(azimuth=azimuth, elevation=elevation, distance=distance)

    def plot_lambda_search(self, lmbds, chisqs, mems, maxcurve):

        ax1 = self.lamda_search_plot.fig.axes[0]

        best_lmbd = lmbds[maxcurve]

        ax1.plot(chisqs, mems, 'ko', ms=self.ms)
        ax1.plot(chisqs[maxcurve], mems[maxcurve], 'ro', ms=self.ms + 1,
                 label='Best lmbd=' + str(best_lmbd))
        ax1.set_xlabel('$\chi^2$', fontsize=self.fs)
        ax1.set_ylabel('Entropy', fontsize=self.fs)

        ax1.tick_params(axis='both', labelsize=self.tls)
        ax1.tick_params(axis='both', labelsize=self.tls)

        ax1.legend()

        self.lamda_search_plot.draw()



import numpy as np
from SpotDIPy import SpotDIPy
import multiprocessing
import pickle
from glob import glob


DIP = SpotDIPy(cpu_num=multiprocessing.cpu_count() - 1, platform_name='cpu')

""" Set required parameters """
DIP.set_param('t0', value=2454254.621769)  # Refernce time [day]
DIP.set_param('period',  value=2.981897)   # Equatorial rotation period [day]
DIP.set_param('Tphot', value=6539)  # Photospheric temperature [K]
DIP.set_param('Tcool', value=4188)  # Minimum spot temperature [K]
DIP.set_param('Thot', value=7500)  # Maximum spot temperature [K]
DIP.set_param('incl', value=45)  # Axial inclination [deg]
# DIP.set_param('vsini', value=20.138)  # Equatorial rotation velocity [km/s]
DIP.set_param('R', value=1.668)  # Stellar radius [Ro]
DIP.set_param('vrt', value=6.79)  # Macroturbulence velocity [km/s]
DIP.set_param('mass', value=1.41)  # Stellar Mass [Mo]
DIP.set_param('dOmega', value=0.0)  # Delta Omega [rad/day] (for differantial rotation)
DIP.set_param('resolution', value=65000)  # Spectral resolution
""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""

""" Set Lim-darkening parameters (via ExoTiC-LD Python package) """
# mh == Metallicity
# law == LD law
# model == Model atmosphere
# mu_min = Minumum mu angle
DIP.set_limb_darkening_params(mh=-0.18, law='linear', model='mps2', mu_min=0.1,
                              data_path='/media/eng/Storage/ExoTiC-LD_data_v3.1.2')
""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""

""" Set modes """
DIP.set_conf({
    'line': {'mode': 'on',  # If observed spectral data is available
             'wave_range': [4412, 7838],  # Spectral wavelength range
             'eqw': 0.08547,  # Equivalent width
             'scaling': {'method': 'mean'},  # Scaling method for observed and synthetic profiles
             'corr': {'rv': None, 'amp': None}  # Radial velocity and amplitude correction
             },
    'lc': {'mode': 'on',  # If observed light curve data is available
           'passband': 'CorotAS',  # Passband of the observed light curve
           'scaling': {'method': 'mean'},  # Scaling method for observed and synthetic light curves
           'corr': {'amp': None}  # Amplitude correction
           }
})
""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""

""" Construction surface grid """
# method = Method of construction surface grid --> healpy, trapezoid, phoebe2_marching
# nlats = Number of latitudes for trapezoid surface grid
# DIP.construct_surface_grid(method='phoebe2_marching', noes=11000)
DIP.construct_surface_grid(method='trapezoid', nlats=40)
""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""

""" Import initial local profiles (LLP) (for photosphere, cool and hot spots) data"""
llp_vels = np.loadtxt('spectrum_Teff6539_logg4.12_mh-0.18_alpha0.06_vmic1.88_vmac0.0_vsini0.0_'
                      'ld0.0_R0_with_err_lsd.out', skiprows=2)[:, 0]
llp_phot_int = np.loadtxt('spectrum_Teff6539_logg4.12_mh-0.18_alpha0.06_vmic1.88_vmac0.0_vsini0.0'
                          '_ld0.0_R0_with_err_lsd.out', skiprows=2)[:, 1]
llp_cool_int = np.loadtxt('spectrum_Teff4188_logg4.12_mh-0.18_alpha0.06_vmic1.88_vmac0.0_vsini0.0'
                          '_ld0.0_R0_with_err_lsd.out', skiprows=2)[:, 1]
llp_hot_int = np.loadtxt('spectrum_Teff7000_logg4.12_mh-0.18_alpha0.06_vmic1.88_vmac0.0_vsini0.0'
                         '_ld0.0_R0_with_err_lsd.out', skiprows=2)[:, 1]

DIP.set_local_profiles({'line': {'lp_vels': llp_vels, 'phot_lp_data': llp_phot_int, 'cool_lp_data': llp_cool_int,
                                 'hot_lp_data': llp_hot_int}})
""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""

""" Make a test """
# lats_spots = [0, 30., 60., -30.]  # spot latitudes (degrees)
# longs_spots = [0., 90., 180., 270.]  # spot longitudes (degrees)
# rs_spots = [15., 15, 15., 15.]  # spot radii (degrees)
# cs_cools = [0.9, 0.0, 0.8, 0.0]  # cool spot contrast between 0 and 1
# cs_hots = [0.0, 0.8, 0.0, 0.8]  # hot spot contrast between 0 and 1
#
# spots_params = {'lats_spots': lats_spots, 'longs_spots': longs_spots, 'rs_spots': rs_spots, 'cs_cools': cs_cools,
#                 'cs_hots': cs_hots}
#
# line_phases = np.arange(0, 1.0, 0.1)
# line_times = DIP.params['t0'] + DIP.params['period'] * line_phases
# line_vels = np.arange(-32, 32 + 1.0, 1.0)
# line_snr = 3000
#
# lc_phases = np.arange(0.0, 1.0 + 0.01, 0.01)
# lc_times = DIP.params['t0'] + DIP.params['period'] * lc_phases
# lc_snr = 3000
#
# opt_params = {'alpha': 1.0, 'beta': 1.0, 'gamma': 1.0, 'delta': 1.0, 'lmbd': 1, 'maxiter': 5000,
#               'tol': 1e-5, 'disp': True}
#
# modes_inp = {'line': {'times': line_times, 'vels': line_vels, 'snr': line_snr},
#              'lc': {'times': lc_times, 'snr': lc_snr}}
#
# plot_params = {'line_sep_prf': 0.05, 'line_sep_res': 0.005, 'mol_sep_prf': 0.03, 'mol_sep_res': 0.01,
#                'show_err_bars': True, 'fmt': '%0.3f', 'markersize': 2, 'linewidth': 1, 'fontsize': 15,
#                'ticklabelsize': 12}
#
# DIP.test(spots_params, modes_input=modes_inp, opt_params=opt_params, plot_params=plot_params,
#          save_data_path="test_find_per_incl.pkl")
""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""

file = open('test_find_per_incl.pkl', 'rb')
input_data_dict = pickle.load(file)
file.close()

# data = np.zeros((len(input_data_dict['lc']['times']), 3))
# data[:, 0] = input_data_dict['lc']['times']
# data[:, 1] = input_data_dict['lc']['data'][:, 0]
# data[:, 2] = input_data_dict['lc']['data'][:, 1]
# np.savetxt('target_lc/Test_lc.txt', data)

# for i, item in enumerate(input_data_dict['line']['times']):
#     data = input_data_dict['line']['data'][i]
#     np.savetxt('target_LSDs/Test_' + str(item) + '_lsd.out', data)

paths = glob("target_LSDs/Test_*.out")
line_times = []
line_data = []
for path in paths:
    line_times.append(float(path.split('/')[-1].split('_')[1]))
    data = np.loadtxt(path)
    line_data.append(data)

data = np.loadtxt("target_lc/Test_lc.txt")
lc_times = data[:, 0]
lc_flux = data[:, 1]
lc_err = data[:, 2]
lc_data = np.vstack((data[:, 1], data[:, 2])).T

input_data_dict = {'line': {'times': line_times, 'data': line_data},
                   'lc': {'times': lc_times, 'data': lc_data}}

DIP.set_input_data(input_data_dict)

""" Grid Search """
# opt_params = {'alpha': 1.0, 'beta': 1.0, 'gamma': 1.0, 'delta': 1.0, 'lmbd': 2.0, 'maxiter': 3500,
#               'tol': 1e-5, 'disp': True}
#
# fit_params = {
#                 # 'dOmega': np.linspace(-1.0, -0.1, 20),
#                 'period': np.linspace(2.5, 3.5, 30),
#                 # 'vsini': np.linspace(16, 23, 10),
#                 # 'R': np.linspace(1.3, 1.8, 20),
#                 # 'vrt': np.linspace(0, 10, 20),
#                 # 'resolution': np.linspace(0.0, 65000, 50),
#                 'incl': np.arange(10, 90, 2.5),
#                 # 'eqw': np.linspace(0.0858 - 0.003, 0.0858 + 0.001, 50)
#                 # 'Tcool': np.linspace(3800, 6539, 30),
#                 # 'Thot': np.linspace(6539, 8500, 30)
#
#                 }
#
# DIP.grid_search(fit_params, opt_params, save='sil2.pkl')
""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""

recons_result = DIP.reconstructor(alpha=1.0, beta=1.0, gamma=1.0, delta=1, lmbd=7, maxiter=5500, tol=1e-5,
                                  disp=True)

DIP.plot(plot_params={'line_sep_prf': 0.06, 'line_sep_res': 0.01, 'mol_sep_prf': 0.3, 'mol_sep_res': 0.2,
                      'show_err_bars': True, 'fmt': '%0.3f', 'markersize': 2, 'linewidth': 1, 'fontsize': 15,
                      'ticklabelsize': 12})


from functools import reduce
from itertools import combinations
import numpy as np
from astropy.coordinates import spherical_to_cartesian as ast_stc
from astropy.coordinates import cartesian_to_spherical as ast_cts
from astropy import units as au, constants as ac
from scipy.interpolate import interp1d, griddata
from scipy.optimize import curve_fit, newton as sc_newton
from scipy.special import roots_legendre
from PyAstronomy import pyasl
from exotic_ld import StellarLimbDarkening
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable
import PyDynamic as PyD
import multiprocessing
import tqdm
from . import cutils
# import cutils


try:
    import healpy as hp
except:
    pass


from PyQt5.QtWidgets import QApplication, QDesktopWidget
app = QApplication([])
screen_info = QDesktopWidget().screenGeometry()
screen_width = screen_info.width()
px = 1 / plt.rcParams['figure.dpi']
figsize = (screen_width * px / 3, screen_width * px / 3)
app.exit()


sol = ac.c.to(au.kilometer / au.second).value  # km/s


def p2m_surface_grid(container, requiv, noes, t0, period, mass):

    import os
    os.environ['PHOEBE_ENABLE_ONLINE_PASSBANDS'] = 'FALSE'
    import phoebe
    from phoebe import u as phoebe_units

    b = phoebe.default_star()

    b.set_value('incl', value=90)
    b.set_value('period', value=period * phoebe_units.day)
    b.set_value('t0', value=t0 * phoebe_units.day)
    b.set_value('requiv', value=requiv * phoebe_units.solRad)
    b.set_value('mass', value=mass * phoebe_units.solMass)

    b['ntriangles'] = noes

    b.add_dataset('mesh', compute_times=t0, coordinates=['xyz'],
                  columns=['xs', 'ys', 'zs', 'areas', 'loggs', 'rs', 'volume'])

    b.run_compute(distortion_method='rotstar', irrad_method='none')

    grid_xyzs = b['mesh@xyz_elements@latest@model'].value.copy()
    grid_xcs = b['mesh@xs@latest@model'].value.copy()
    grid_ycs = b['mesh@ys@latest@model'].value.copy()
    grid_zcs = b['mesh@zs@latest@model'].value.copy()
    grid_areas = b['mesh@areas@latest@model'].value.copy()
    grid_loggs = b['mesh@loggs@latest@model'].value.copy()

    rs, lats, longs = ast_cts(grid_xcs, grid_ycs, grid_zcs)

    grid_rs = rs.copy()
    grid_lats = lats.value.copy()
    grid_longs = longs.value.copy()

    container['method'] = 'phoebe2_marching'
    container['grid_xyzs'] = grid_xyzs
    container['grid_areas'] = grid_areas
    container['grid_loggs'] = grid_loggs
    container['grid_lats'] = grid_lats
    container['grid_longs'] = grid_longs
    container['grid_rs'] = grid_rs


def hp_surface_grid(container, omega, nside, radius, mass, cpu_num=1):

    npix = hp.nside2npix(nside)

    input_args = [(nside, pixind, omega, radius) for pixind in range(npix)]
    if cpu_num > 1:
        pool = multiprocessing.Pool(processes=cpu_num)
        results = pool.map(hp_surface_grid_main, input_args)
        pool.close()
        pool.join()
    else:
        results = []
        for item in input_args:
            results.append(hp_surface_grid_main(item))

    parts = np.array(results, dtype=object)

    grid_areas = np.hstack(parts[:, 0])
    grid_xcs = np.hstack(parts[:, 1])
    grid_ycs = np.hstack(parts[:, 2])
    grid_zcs = np.hstack(parts[:, 3])
    grid_xyzs = np.vstack(parts[:, 4])

    rs, lats, longs = ast_cts(grid_xcs, grid_ycs, grid_zcs)

    grid_rs = rs.value.copy()
    grid_lats = lats.value.copy()
    grid_longs = longs.value.copy()
    grid_loggs = calc_logg(mass=mass, re=radius, rs=grid_rs, omega=omega, theta=grid_lats + np.pi / 2.0)

    container['method'] = 'healpy'
    container['grid_xyzs'] = grid_xyzs
    container['grid_areas'] = grid_areas
    container['grid_loggs'] = grid_loggs
    container['grid_lats'] = grid_lats
    container['grid_longs'] = grid_longs
    container['grid_rs'] = grid_rs


def hp_surface_grid_main(args):

    nside, pixind, omega, radius = args

    center = hp.pix2vec(nside, pixind, nest=False)
    polygon = hp.boundaries(nside, pixind, step=1, nest=False)

    _, plats, plongs = ast_cts(polygon[0], polygon[1], polygon[2])

    xps = []
    yps = []
    zps = []
    for plat, plong in zip(plats.value, plongs.value):
        rplat = calc_r_to_Re_ratio(omega=omega, theta=plat + np.pi / 2.0)
        rplat = rplat * radius

        xp, yp, zp = ast_stc(rplat, plat, plong)

        xps.append(xp.value)
        yps.append(yp.value)
        zps.append(zp.value)

    polygon = [[[xps[0], yps[0], zps[0]],
                [xps[1], yps[1], zps[1]],
                [xps[2], yps[2], zps[2]]],

               [[xps[2], yps[2], zps[2]],
                [xps[3], yps[3], zps[3]],
                [xps[0], yps[0], zps[0]]]]

    cr, clat, clong = ast_cts(center[0], center[1], center[2])

    rclat = calc_r_to_Re_ratio(omega=omega, theta=clat.value + np.pi / 2.0)
    rclat = rclat * radius

    xpc, ypc, zpc = ast_stc(rclat, clat.value, clong.value)

    area = calc_triangle_area(polygon[0]) + calc_triangle_area(polygon[1])

    return area, xpc.value, ypc.value, zpc.value, polygon


def td_surface_grid(container, omega, nlats, radius, mass, cpu_num=1):

    blats = np.linspace(np.pi / 2., -np.pi / 2., nlats + 1)

    ipn = 4
    pzh = abs(np.sin(blats[0]) - np.sin(blats[1])) * radius
    uarea = 2. * np.pi * radius * pzh / ipn

    input_args = [([blats[i], blats[i + 1]], uarea, omega, radius, mass) for i in range(len(blats) - 1)]

    if cpu_num > 1:
        pool = multiprocessing.Pool(processes=cpu_num)
        # results = pool.map(td_surface_grid_main, input_args)
        results = pool.map(cutils.td_grid_sect, input_args)
        pool.close()
        pool.join()
    else:
        results = []
        for item in input_args:
            results.append(cutils.td_grid_sect(item))

    parts = np.array(list(results), dtype=object)

    container['method'] = 'trapezoid'
    container['blats'] = blats
    container['nlons'] = parts[:, 0]
    container['blongs'] = parts[:, 1]
    container['grid_areas'] = np.hstack(parts[:, 2]).reshape(-1, 2).sum(axis=1)
    container['grid_rs'] = np.hstack(parts[:, 3])
    container['grid_lats'] = np.hstack(parts[:, 4])
    container['grid_longs'] = np.hstack(parts[:, 5])
    container['grid_xyzs'] = np.vstack(parts[:, 6])
    container['grid_loggs'] = np.hstack(parts[:, 7])


# def td_surface_grid_main(args):
#
#     blatpn, uarea, omega, radius, mass = args
#
#     zh = abs(np.sin(blatpn[0]) - np.sin(blatpn[1])) * radius
#     zarea = 2. * np.pi * radius * zh
#
#     nlon = int(round(zarea / uarea))
#     blong = np.linspace(0.0, 2.0 * np.pi, nlon + 1)
#
#     areas = []
#     uvws = []
#
#     inds1 = [(0, 0), (1, 0), (0, 1)]
#     inds2 = [(1, 0), (1, 1), (0, 1)]
#     for j in range(nlon):
#         for inds in [inds1, inds2]:
#             polygon = []
#
#             for ind in inds:
#                 rlat = calc_r_to_Re_ratio(omega=omega, theta=blatpn[ind[0]] + np.pi / 2.0)
#                 rlat = rlat * radius
#
#                 xp, yp, zp = ast_stc(rlat, blatpn[ind[0]], blong[j + ind[1]])
#
#                 polygon.append([xp.value, yp.value, zp.value])
#             uvws.append(np.array(polygon.copy()))
#
#             area = calc_triangle_area(polygon)
#             areas.append(area)
#
#     rzlat = calc_r_to_Re_ratio(omega=omega, theta=(blatpn[0] + blatpn[1] + np.pi) / 2.0)
#     rzlat = rzlat * radius
#
#     xpc, ypc, zpc = ast_stc(rzlat, [(blatpn[0] + blatpn[1]) / 2.0] * nlon,
#                             (blong[1:] + blong[:-1]) / 2.)
#
#     rs, lats, longs = ast_cts(xpc, ypc, zpc)
#
#     loggs = calc_logg(mass=mass, re=radius, rs=rs.value, omega=omega, theta=lats.value + np.pi / 2.0)
#
#     return nlon, blong, areas, rs.value, lats.value, longs.value, uvws, loggs


def calc_triangle_area(polygon):

    v1 = np.array(polygon[0])
    v2 = np.array(polygon[1])
    v3 = np.array(polygon[2])

    vector1 = v2 - v1
    vector2 = v3 - v1

    cross_product = np.cross(vector1, vector2)

    area = np.linalg.norm(cross_product) / 2.0

    return area


def input_data_prep(container, conf):

    scale_factor = 1.0
    max_reg = [1.0]

    for mode in conf:
        if mode != 'lc':
            times = np.array(container[mode]['times'])
            data = np.array(container[mode]['data'])

            srt = np.argsort(times)
            times = times[srt]
            data = data[srt]

            data_dict = {}
            data_cube = [[], []]
            for i, itime in enumerate(times):

                data_dict[itime] = {}

                ovels, oprf, oerrs = data[i][:, 0], data[i][:, 1], data[i][:, 2]

                if conf[mode]['scaling']['method'] == 'mean':
                    scale_factor = np.average(oprf)

                elif conf[mode]['scaling']['method'] == 'max':
                    per = 100. / conf[mode]['scaling']['percent']
                    if conf[mode]['scaling']['side'] == 'both':
                        max_reg = np.hstack((oprf[:int(len(oprf) / per)], oprf[-int(len(oprf) / per):]))
                    elif conf[mode]['scaling']['side'] == 'left':
                        max_reg = oprf[:int(len(oprf) / per)]
                    elif conf[mode]['scaling']['side'] == 'right':
                        max_reg = oprf[-int(len(oprf) / per):]
                    scale_factor = np.average(max_reg)

                elif conf[mode]['scaling']['method'] == 'region':
                    owavs = ((ovels * np.average(conf[mode]['wrange'])) / sol + np.average(conf[mode]['wrange']))

                    wavgw = []
                    wavgf = []
                    for region in conf[mode]['scaling']['ranges']:
                        wr = np.argwhere((owavs >= region[0]) & (owavs <= region[1])).T[0]

                        wavgw.append(np.average(owavs[wr]))
                        wavgf.append(np.average(oprf[wr]))

                    x1, x2, x3 = wavgw
                    y1, y2, y3 = wavgf
                    denom = (x1 - x2) * (x1 - x3) * (x2 - x3)
                    A = (x3 * (y2 - y1) + x2 * (y1 - y3) + x1 * (y3 - y2)) / denom
                    B = (x3 * x3 * (y1 - y2) + x2 * x2 * (y3 - y1) + x1 * x1 * (y2 - y3)) / denom
                    C = (x2 * x3 * (x2 - x3) * y1 + x3 * x1 * (x3 - x1) * y2 + x1 * x2 * (x1 - x2) * y3) / denom
                    scale_factor = A * owavs ** 2 + B * owavs + C

                oprf /= scale_factor
                oerrs /= scale_factor

                if container[mode]['vels'] is not None:
                    ovels, oprf, oerrs = PyD.interp1d_unc(container[mode]['vels'], ovels, oprf, oerrs, kind='linear')

                data_dict[itime]['vels'] = ovels.copy()
                data_dict[itime]['prf'] = oprf.copy()
                data_dict[itime]['errs'] = oerrs.copy()
                data_cube[0].append(oprf.copy())
                data_cube[1].append(oerrs.copy())

            container[mode]['times'] = times.copy()
            container[mode]['data'] = data_dict.copy()
            container[mode]['vels'] = data_dict[times[0]]['vels'].copy()
            container[mode]['noo'] = len(container[mode]['times'])
            container[mode]['nop'] = len(container[mode]['vels'])
            container[mode]['data_cube'] = np.array(data_cube)

        else:
            times = np.array(container[mode]['times'])
            data = np.array(container[mode]['data'])
            fluxs = data[:, 0]
            errs = data[:, 1]

            srt = np.argsort(times)
            times = times[srt]
            fluxs = fluxs[srt]
            errs = errs[srt]

            if container[mode]['norsp'] is not None:
                norsp = container[mode]['norsp']
                ntimes = np.linspace(min(times), max(times), norsp)
                times, fluxs, errs = PyD.interp1d_unc(ntimes, times, fluxs, errs, kind='linear')

            if conf[mode]['scaling']['method'] == 'mean':
                scale_factor = np.average(fluxs)
            fluxs /= scale_factor
            errs /= scale_factor

            data_dict = {'fluxs': fluxs, 'errs': errs}

            container[mode]['times'] = times.copy()
            container[mode]['data'] = data_dict.copy()
            container[mode]['noo'] = 1
            container[mode]['nop'] = len(container[mode]['times'])
            container[mode]['data_cube'] = np.array([fluxs, errs])


def get_total_fs(fssc, fssh, areas, lats, incl):

    total_fsc = np.sum(fssc * areas) / np.sum(areas)
    total_fsh = np.sum(fssh * areas) / np.sum(areas)
    total_fsp = np.sum((1. - fssc - fssh) * areas) / np.sum(areas)

    w = np.argwhere(lats >= -np.deg2rad(incl)).T[0]
    partial_fsc = np.sum(fssc[w] * areas[w]) / np.sum(areas[w])
    partial_fsh = np.sum(fssh[w] * areas[w]) / np.sum(areas[w])
    partial_fsp = np.sum((1. - fssc[w] - fssh[w]) * areas[w]) / np.sum(areas[w])

    return total_fsc, total_fsh, total_fsp, partial_fsc, partial_fsh, partial_fsp


def calc_logg(mass, re, rs, omega, theta):

    mass_kg = mass * au.solMass.to(au.kg)
    g_cm = mass_kg * ac.G.to(au.cm**3 / (au.kg * au.s**2)).value
    r_cm = re * au.solRad.to(au.cm)
    r_ = rs / re

    if type(theta) in [list, np.ndarray]:
        theta[theta > np.pi / 2.] = np.pi - theta[theta > np.pi / 2.]

    else:
        theta = np.pi - theta if theta > np.pi / 2. else theta

    geff = (g_cm / r_cm**2) * (1.0 / r_**4 + omega**4 * r_**2 * np.sin(theta)**2
                               - 2.0 * omega**2 * np.sin(theta)**2 / r_)**0.5

    return np.log10(geff)


def calc_gds(omega, thetas):

    gds = np.zeros(len(thetas))
    for i, theta in enumerate(thetas):

        theta = np.pi - theta if theta > np.pi / 2. else theta

        r = calc_r_to_Re_ratio(omega=omega, theta=theta)

        if theta == 0:
            fluxw = np.exp(2.0 * omega**2 * r**3 / 3.0)

        elif theta == np.pi / 2.:
            fluxw = (1.0 - omega**2 * r**3)**-(2.0 / 3.0)

        else:
            nu = calc_nu(omega=omega, r=r, theta=theta)
            fluxw = np.tan(nu)**2 / np.tan(theta)**2

        gds[i] = ((r**-4 + omega**4 * r**2 * np.sin(theta)**2 - (2.0 * omega**2 * np.sin(theta)**2 / r))**(1.0/8.0)
                  * fluxw**0.25)

    return gds


def r_func(r, omega, theta):

    return omega ** -2 + 0.5 - (omega ** 2 * r) ** -1 - 0.5 * r ** 2 * np.sin(theta)**2


def calc_r_to_Re_ratio(omega, theta):

    theta = np.pi - theta if theta > np.pi / 2. else theta
    r = sc_newton(r_func, 1.0, args=(omega, theta))

    return r


def nu_func(nu, omega, r, theta):

    return (np.cos(nu) + np.log(np.tan(nu / 2.)) - omega**2 * r**3 *
            np.cos(theta)**3 / 3.0 - np.cos(theta) - np.log(np.tan(theta / 2.)))


def calc_nu(omega, r, theta):

    theta = np.pi - theta if theta > np.pi / 2. else theta

    nu = sc_newton(nu_func, theta, args=(omega, r, theta))

    return nu


def calc_radius(vsini, incl, period):

    veq = vsini / np.sin(np.deg2rad(incl))
    radius = (veq * period * 86400. / (2. * np.pi)) * au.km.to(au.solRad)

    return radius


def calc_omega_and_requiv(mass, period, re):

    mass_kg = mass * au.solMass.to(au.kg)
    gm = mass_kg * ac.G.value
    period_s = period * au.day.to(au.second)
    re_m = re * au.solRad.to(au.m)
    omega = (2.0 * np.pi / period_s) * np.sqrt(re_m ** 3 / gm)
    rp = calc_r_to_Re_ratio(omega=omega, theta=0) * re

    requiv = (re ** 2 * rp) ** (1. / 3.)

    return omega, requiv, rp


def gaussian(x, amp, cen, sig, lvl):
    return amp * np.exp(-(x - cen) ** 2 / (2 * sig ** 2)) + lvl


def fit_gaussian(x, y):
    cen = np.sum(x * y) / np.sum(y)
    sig = np.sqrt(np.sum(y * (x - cen) ** 2) / np.sum(y))
    amp, lvl = max(y), min(y)

    bounds = [(-np.inf, -np.inf, 0.0, -np.inf), (np.inf, np.inf, np.inf, np.inf)]
    popt, _ = curve_fit(gaussian, x, y, p0=[amp, cen, sig, lvl], bounds=bounds)

    return popt


def set_eqw(vels, ints, wrange, eqw):

    amp, cen, sig, lvl = fit_gaussian(vels, ints)
    cent_wave = (wrange[0] + wrange[1]) / 2.
    ceqw = (amp * sig * np.sqrt(2 * np.pi)) / sol * cent_wave * -1.0
    median = np.median(ints)
    neqw = eqw / (ceqw * median)

    return (ints - median) * neqw + 1.0


def set_instrbroad(wrange, vels, ints, resolution):

    wb = np.average(wrange)
    waves = wb + vels * wb / sol

    convolved_ints = pyasl.instrBroadGaussFast(waves, ints, resolution, edgeHandling="firstlast", fullout=False,
                                               maxsig=None, equid=False)

    return convolved_ints


def _integrate_I_mu(sc, wavelength_range, mode, custom_wavelengths=None, custom_throughput=None):

    """ taken from the exotic_ld Python package for the calculation of Imu_0 """

    if mode == "custom":
        s_wavelengths = custom_wavelengths
        s_throughputs = custom_throughput
    else:
        s_wavelengths, s_throughputs = sc._read_sensitivity_data(mode)

    wavelength_range = np.sort(np.array(wavelength_range))
    s_mask = np.logical_and(wavelength_range[0] < s_wavelengths,
                            s_wavelengths < wavelength_range[1])

    i_mask = np.logical_and(wavelength_range[0] < sc.stellar_wavelengths,
                            sc.stellar_wavelengths < wavelength_range[1])

    s_wvs = s_wavelengths[s_mask]
    s_thp = s_throughputs[s_mask]
    i_wvs = sc.stellar_wavelengths[i_mask]
    i_int = sc.stellar_intensities[i_mask]

    # Ready sensitivity interpolator.
    if s_wvs.shape[0] >= 2:
        s_interp_func = interp1d(s_wvs, s_thp, kind="linear", bounds_error=False, fill_value=0.)
    else:
        mean_wv = np.mean(wavelength_range)
        match_wv_idx = np.argmin(np.abs(s_wavelengths - mean_wv))
        match_s = s_throughputs[match_wv_idx]
        s_interp_func = lambda _sw: np.ones(_sw.shape) * match_s

    # Pre-compute Gauss-legendre roots and rescale to lims.
    roots, weights = roots_legendre(500)
    a = wavelength_range[0]
    b = wavelength_range[1]
    t = (b - a) / 2 * roots + (a + b) / 2

    # Iterate mu values computing intensity.
    I_mu = np.zeros(i_int.shape[1])
    for mu_idx in range(sc.mus.shape[0]):

        # Ready intensity interpolator.
        if i_wvs.shape[0] >= 2:
            i_interp_func = interp1d(
                i_wvs, i_int[:, mu_idx], kind="linear",
                bounds_error=False, fill_value=0.)
        else:
            mean_wv = np.mean(wavelength_range)
            match_wv_idx = np.argmin(np.abs(sc.stellar_wavelengths - mean_wv))
            match_i = sc.stellar_intensities[match_wv_idx, mu_idx]
            i_interp_func = lambda _iw: np.ones(_iw.shape) * match_i

        def integrand(_lambda):
            return s_interp_func(_lambda) * i_interp_func(_lambda)

        # Approximate integral.
        I_mu[mu_idx] = (b - a) / 2. * integrand(t).dot(weights)

    return I_mu[0]


def _get_stellar_model_kd_tree(ld_model, ld_data_version=""):
    """ taken from the exotic_ld Python package for the download required model atmospehere files before
    multiprocessing """

    import pkg_resources
    import pickle

    tree_path = pkg_resources.resource_filename(
        "grid_build.kd_trees", "{}_tree{}.pickle".format(
            ld_model, ld_data_version))

    try:
        with open(tree_path, "rb") as f:
            stellar_kd_tree = pickle.load(f)
    except FileNotFoundError as err:
        raise ValueError("ld_model not recognised.")
    except pickle.UnpicklingError as err:
        raise pickle.UnpicklingError("Failed loading stellar model pickle, "
                                     "check python version is supported.")

    return stellar_kd_tree


def ldcs_rgi_prep(teffps, teffcs, teffhs, loggs, law, mh, model, data_path, mu_min, wrange=None, passband=None,
                  cpu_num=1):
    nop = len(teffps)
    teffs = np.hstack((teffps, teffcs, teffhs))
    loggs = np.array(list(loggs) * 3)

    r_M_H = 1.00
    r_Teff = 607.
    r_logg = 1.54
    par_grid = np.vstack(([mh / r_M_H] * len(teffs), teffs / r_Teff, loggs / r_logg)).T

    stellar_kd_tree = _get_stellar_model_kd_tree(ld_model=model)
    distance, nearest_idx = stellar_kd_tree.query(par_grid, k=1)
    matches = stellar_kd_tree.data[nearest_idx] * np.array([r_M_H, r_Teff, r_logg])
    umatches = np.unique(matches, axis=0)

    input_args = []
    for order, item in enumerate(umatches):
        input_args += [(item[1], item[2], law, mh, model, data_path, mu_min, wrange, passband, order)]

    if cpu_num > 1:
        pool = multiprocessing.Pool(processes=cpu_num)
        results = pool.map(ldcs_rgi_prep_main, input_args)
        pool.close()
        pool.join()
    else:
        results = []
        for input_arg in input_args:
            results.append(ldcs_rgi_prep_main(input_arg))

    parts = np.array(results, dtype=object)
    srt = np.argsort(parts[:, -1])
    parts = parts[srt]

    ldc1 = np.zeros(len(teffs))
    ldc2 = np.zeros(len(teffs))
    Imu0 = np.zeros(len(teffs))
    for i, x_row in enumerate(matches):
        for j, y_row in enumerate(umatches):
            if np.array_equal(x_row, y_row):
                ldc1[i] = parts[j, 0]
                ldc2[i] = parts[j, 1]
                Imu0[i] = parts[j, 2]

    """ buraya dikkat. intensiti değerleri cok yüksek olduğu için phot sıcaklığı ve en küçük logg (ekvatora yakın diye)
     ye karşılık gelen Imu0 a normalize edildi. böylece ışık eğrisi çok yüksek değerlerde oluşmayacak ve amp free
      olduğunda uygun değere gidebilecek"""
    # base_Imu0 = ldcs_rgi_prep_main([teffs[0], np.min(loggs), law, mh, model, data_path, mu_min, wrange, passband, 0])[-2]
    # Imu0 /= base_Imu0
    Imu0 /= np.mean(Imu0)
    """"""

    rgi = {'phot': [ldc1[:nop], ldc2[:nop], Imu0[:nop]],
           'cool': [ldc1[nop:nop * 2], ldc2[nop:nop * 2], Imu0[nop:nop * 2]],
           'hot': [ldc1[nop * 2:nop * 3], ldc2[nop * 2:nop * 3], Imu0[nop * 2:nop * 3]]}

    return rgi


def ldcs_rgi_prep_main(args):
    temp, logg, law, mh, model, data_path, mu_min, wrange, passband, order = args

    if wrange is not None:
        ldc, I0 = ld_coeff_calc(ld_law=law, teff=temp, logg=logg, mh=mh, model=model,
                                data_path=data_path, mu_min=mu_min, wrange=wrange)

    elif passband is not None:
        ldc, I0 = ld_coeff_calc(ld_law=law, teff=temp, logg=logg, mh=mh, model=model,
                                data_path=data_path, mu_min=mu_min, passband=passband)

    if law == 'linear':
        return ldc[0], 0.0, I0, order

    else:
        return ldc[0], ldc[1], I0, order


def ld_coeff_calc(ld_law, teff, logg, mh, model, data_path, mu_min=0.1, wrange=None, passband=None):

    sc = StellarLimbDarkening(Teff=teff, logg=logg, M_H=mh, ld_model=model,  # kurucz mps1 mps2 stagger
                              ld_data_path=data_path)

    if ld_law == 'linear':
        clc = sc.compute_linear_ld_coeffs

    if ld_law == 'square-root':
        clc = sc.compute_squareroot_ld_coeffs

    if ld_law == 'quadratic':
        clc = sc.compute_quadratic_ld_coeffs

    if wrange is not None:
        custom_wavelengths, custom_throughput = np.array([np.min(wrange), np.max(wrange)]), np.array([1, 1])

    elif passband is not None:
        custom_wavelengths, custom_throughput = sc._read_sensitivity_data(passband)

    ldc = clc(custom_wavelengths, mode='custom', custom_wavelengths=custom_wavelengths,
              custom_throughput=custom_throughput, mu_min=mu_min, return_sigmas=False)

    I_mu0 = _integrate_I_mu(sc=sc, wavelength_range=custom_wavelengths, mode='custom',
                            custom_wavelengths=custom_wavelengths, custom_throughput=custom_throughput)

    return ldc, I_mu0


def generate_spotted_surface(surface_grid, spots_params, default=1e-5):

    lats_spots = spots_params['lats_spots']
    longs_spots = spots_params['longs_spots']
    rs_spots = spots_params['rs_spots']
    cs_cools = spots_params['cs_cools']
    cs_hots = spots_params['cs_hots']

    cmap = np.ones(len(surface_grid['grid_lats'])) * default
    hmap = np.ones(len(surface_grid['grid_lats'])) * default
    for (lat_spot, long_spot, r_spot, c_cool, c_hot) in zip(lats_spots, longs_spots, rs_spots, cs_cools, cs_hots):

        dlon = surface_grid['grid_longs'] - np.deg2rad(long_spot)
        angles = np.arccos(np.sin(surface_grid['grid_lats']) * np.sin(np.deg2rad(lat_spot)) +
                           np.cos(surface_grid['grid_lats']) * np.cos(np.deg2rad(lat_spot)) * np.cos(dlon))

        ii = np.argwhere(angles <= np.deg2rad(r_spot)).T[0]
        ni = len(ii)
        if ni > 0:
            cmap[ii] = c_cool
            hmap[ii] = c_hot

    return cmap, hmap


def td_to_rect_map(lats, longs, ints, xsize, ysize):

    xlats = np.linspace(np.pi / 2., -np.pi / 2., ysize)
    xlongs = np.linspace(0, 2.0 * np.pi, xsize)

    mlongs, mlats = np.meshgrid(xlongs, xlats)

    points = np.array([lats, longs]).T
    grid_points = (mlats, mlongs)

    rmap = griddata(points, ints, grid_points, method='nearest')

    xs = (max(xlongs) - min(xlongs)) / (len(xlongs) - 1) / 2.
    ys = (max(xlats) - min(xlats)) / (len(xlats) - 1) / 2.

    extent = [min(xlongs) - xs, max(xlongs) + xs, min(xlats) - ys, max(xlats) + ys]

    return xlongs, xlats, rmap, extent, mlats, mlongs


def hp_to_rect_map(ints, nside, xsize=180, ysize=90):

    xlats = np.linspace(0.0, np.pi, ysize)
    xlongs = np.linspace(0.0, 2 * np.pi, xsize)

    mlongs, mlats = np.meshgrid(xlongs, xlats)
    grid_pix = hp.ang2pix(nside, mlats, mlongs, nest=False)

    rmap = ints[grid_pix]

    xlats = xlats[::-1] - np.pi / 2.
    xs = (max(xlongs) - min(xlongs)) / (len(xlongs) - 1) / 2.
    ys = (max(xlats) - min(xlats)) / (len(xlats) - 1) / 2.

    extent = [min(xlongs) - xs, max(xlongs) + xs, min(xlats) - ys, max(xlats) + ys]

    return xlongs, xlats, rmap, extent, np.flipud(mlats - np.pi / 2.), mlongs


def p2m_to_rect_map(ints, grid_xyzs, xsize, ysize, subno=2):

    _, rd_ints, _, lats, longs = cutils.subdivision(grid_xyzs, ints, subno=subno)

    xlats = np.linspace(np.pi / 2., -np.pi / 2., ysize)
    xlongs = np.linspace(0, 2 * np.pi, xsize)

    mlongs, mlats = np.meshgrid(xlongs, xlats)

    points = np.array([lats, longs]).T
    grid_points = (mlats, mlongs)

    rmap = griddata(points, rd_ints, grid_points, method='nearest')

    xs = (max(xlongs) - min(xlongs)) / (len(xlongs) - 1) / 2.
    ys = (max(xlats) - min(xlats)) / (len(xlats) - 1) / 2.

    extent = [min(xlongs) - xs, max(xlongs) + xs, min(xlats) - ys, max(xlats) + ys]

    return xlongs, xlats, rmap, extent, mlats, mlongs


def grid_to_rect_map(surface_grid, ints):

    if surface_grid['method'] == 'trapezoid':

        nlons = surface_grid['nlons'].copy()

        xlongs, xlats, rmap, extent, mlats, mlongs = td_to_rect_map(lats=surface_grid['grid_lats'],
                                                                    longs=surface_grid['grid_longs'], ints=ints,
                                                                    xsize=max(nlons), ysize=surface_grid['nlats'])

    elif surface_grid['method'] == 'healpy':

        xlongs, xlats, rmap, extent, mlats, mlongs = hp_to_rect_map(ints=ints, nside=surface_grid['nside'], xsize=180,
                                                                    ysize=90)

    elif surface_grid['method'] == 'phoebe2_marching':

        xlongs, xlats, rmap, extent, mlats, mlongs = p2m_to_rect_map(ints=ints, grid_xyzs=surface_grid['grid_xyzs'],
                                                                     xsize=180, ysize=90, subno=2)

    mapprojs = {'xlongs': xlongs, 'xlats': xlats, 'rmap': rmap, 'extent': extent, 'mlats': mlats, 'mlongs': mlongs}

    return mapprojs


def test_prf_plot(DIP, mode, plotp):

    times = DIP.idc[mode]['times']
    vels = DIP.idc[mode]['vels']
    data = DIP.idc[mode]['data']
    spotless_slps = DIP.opt_results[mode]['spotless_sprfs']
    recons_slps = DIP.opt_results[mode]['recons_sprfs']
    epochs = (times - DIP.params['t0']) / DIP.params['period']

    plt.figure(figsize=figsize)
    ax3 = plt.subplot2grid((20, 2), (0, 0), colspan=2)
    ax1 = plt.subplot2grid((20, 2), (1, 0), rowspan=19)
    ax2 = plt.subplot2grid((20, 2), (1, 1), rowspan=19)

    ax3.axis("off")

    if mode == 'line':
        plotp['sep_prf'] = plotp['line_sep_prf']
        plotp['sep_res'] = plotp['line_sep_res']

    if mode == 'mol':
        plotp['sep_prf'] = plotp['mol_sep_prf']
        plotp['sep_res'] = plotp['mol_sep_res']

    for i, itime in enumerate(sorted(times)):

        maxv = max(vels)
        maxi = max(data[itime]['prf']) + i * plotp['sep_prf']
        residual = data[itime]['prf'] - recons_slps[itime]['prf']
        maxir = np.average(residual + i * plotp['sep_res'])

        if plotp['show_err_bars']:
            ax1.errorbar(vels, data[itime]['prf'] + i * plotp['sep_prf'],
                         yerr=data[itime]['errs'], fmt='o', color='k', ms=plotp['markersize'])
            ax2.errorbar(vels, residual + i * plotp['sep_res'],
                         yerr=data[itime]['errs'], fmt='o', color='k', ms=plotp['markersize'])
        else:
            ax1.plot(vels, data[itime]['prf'] + i * plotp['sep_prf'], 'ko', ms=plotp['markersize'])
            ax2.plot(vels, residual + i * plotp['sep_res'], 'ko', ms=plotp['markersize'])

        ax1.plot(vels, spotless_slps[itime]['prf'] + i * plotp['sep_prf'], 'b', linewidth=plotp['linewidth'], zorder=2)
        ax1.plot(vels, recons_slps[itime]['prf'] + i * plotp['sep_prf'], 'r', linewidth=plotp['linewidth'], zorder=3)
        ax1.annotate(str('%0.3f' % round(epochs[i], 3)), xy=(maxv - maxv / 3.1, maxi + plotp['sep_prf'] / 10.),
                     color='g')
        ax2.annotate(str('%0.3f' % round(epochs[i], 3)), xy=(maxv - maxv / 10., maxir + plotp['sep_res'] / 10.),
                     color='g')
        ax2.axhline(i * plotp['sep_res'], color='r', zorder=3)

    ax1.set_xlabel('Radial Velocity (km / s)', fontsize=plotp['fontsize'])
    ax1.set_ylabel('I / Ic', fontsize=plotp['fontsize'])

    ax3.plot([], [], 'ko', label='Obs. Data', ms=plotp['markersize'])
    ax3.plot([], [], 'b', label='Spotless Model', linewidth=plotp['linewidth'])
    ax3.plot([], [], 'r', label='Spotted Model', linewidth=plotp['linewidth'])
    ax3.legend(loc='center', ncol=3, frameon=False)

    ax2.set_xlabel('Radial Velocity (km / s)', fontsize=plotp['fontsize'])
    ax2.set_ylabel('Residuals', fontsize=plotp['fontsize'])

    ax1.tick_params(axis='both', labelsize=plotp['ticklabelsize'])
    ax2.tick_params(axis='both', labelsize=plotp['ticklabelsize'])

    plt.tight_layout()


def test_lc_plot(DIP, plotp):

    recons_slc = DIP.opt_results['lc']['recons_slc']

    plt.figure(figsize=figsize)
    ax1 = plt.subplot2grid((3, 1), (0, 0), rowspan=2)
    ax2 = plt.subplot2grid((3, 1), (2, 0))

    if 'lc' in DIP.conf:

        times = DIP.idc['lc']['times']
        fluxs = DIP.idc['lc']['data']['fluxs']
        errs = DIP.idc['lc']['data']['errs']

        epochs = (times - DIP.params['t0']) / DIP.params['period']

        ax1.errorbar(epochs, fluxs, yerr=errs, fmt='o', color='k', label='Obs. Light Curve', ms=plotp['markersize'])
        ax1.plot(epochs, recons_slc, 'r', label='Spotted Model', linewidth=plotp['linewidth'])
        ax2.errorbar(epochs, fluxs - recons_slc, yerr=errs, fmt='o', color='k', ms=plotp['markersize'])
    else:

        ntimes = DIP.opt_results['lc']['ntimes']
        nepochs = (ntimes - DIP.params['t0']) / DIP.params['period']

        ax1.plot(nepochs, recons_slc, 'ko', label='Synthetic Light Curve', ms=plotp['markersize'])
        ax2.set_xlim([0, 2])
        ax2.set_ylim([-1, 1])

    ax1.set_xlabel('Epoch', fontsize=plotp['fontsize'])
    ax1.set_ylabel('Normalized Flux', fontsize=plotp['fontsize'])
    ax1.tick_params(axis='both', labelsize=plotp['ticklabelsize'])

    ax1.legend(loc='upper center', bbox_to_anchor=(0.5, 1.10), ncol=2, frameon=False)

    ax2.axhline(0.0, color='r', linewidth=plotp['linewidth'])

    ax2.set_xlabel('Epoch', fontsize=plotp['fontsize'])
    ax2.set_ylabel('Residuals', fontsize=plotp['fontsize'])
    ax2.tick_params(axis='both', labelsize=plotp['ticklabelsize'])

    plt.tight_layout()


def test_map_plot(DIP, fake_fssc, fake_fssh, fake_total_fs, recons_total_fs, plotp):

    Thot = DIP.params['Thot']
    Tphot = DIP.params['Tphot']
    Tcool = DIP.params['Tcool']
    surface_grid = DIP.surface_grid
    recons_fssc = DIP.opt_results['recons_fssc']
    recons_fssh = DIP.opt_results['recons_fssh']

    fig = plt.figure(figsize=figsize)
    ax1 = plt.subplot2grid((2, 1), (0, 0))
    ax2 = plt.subplot2grid((2, 1), (1, 0))

    fake_fss = (fake_fssc * DIP.params['Tcool'] ** 4 + fake_fssh * DIP.params['Thot'] ** 4 +
                (1.0 - (fake_fssc + fake_fssh)) * DIP.params['Tphot'] ** 4) / DIP.params['Tphot'] ** 4

    recons_fss = (recons_fssc * DIP.params['Tcool'] ** 4 + recons_fssh * DIP.params['Thot'] ** 4 +
                  (1.0 - (recons_fssc + recons_fssh)) * DIP.params['Tphot'] ** 4) / DIP.params['Tphot'] ** 4

    fmapprojs = grid_to_rect_map(surface_grid=DIP.surface_grid, ints=fake_fss)
    cmapprojs = grid_to_rect_map(surface_grid=DIP.surface_grid, ints=recons_fss)

    frmap, fextent = fmapprojs['rmap'], fmapprojs['extent']
    crmap, cextent = cmapprojs['rmap'], cmapprojs['extent']

    vmin = (Tcool / Tphot) ** 4
    vmax = (Thot / Tphot) ** 4

    img2 = ax1.imshow(frmap, cmap='gray', aspect='auto', extent=np.rad2deg(fextent), interpolation='bicubic',
                      vmin=vmin, vmax=vmax)
    img3 = ax2.imshow(crmap, cmap='gray', aspect='auto', extent=np.rad2deg(cextent), interpolation='bicubic',
                      vmin=vmin, vmax=vmax)

    ax1.text(30, -70, 'Total Spotted Area (%) = ' +
             str('%0.3f' % round((fake_total_fs[0] + fake_total_fs[1]) * 100, 3)))
    ax2.text(30, -70, 'Total Spotted Area (%) = ' +
             str('%0.3f' % round((recons_total_fs[0] + recons_total_fs[1]) * 100, 3)))

    divider2 = make_axes_locatable(ax1)
    cax2 = divider2.append_axes('right', size='5%', pad='5%')
    clb2 = fig.colorbar(img2, cax=cax2)
    clb2.set_label(r'$\frac{I}{I_{phot}}$', fontsize=plotp['fontsize'])

    divider3 = make_axes_locatable(ax2)
    cax3 = divider3.append_axes('right', size='5%', pad='5%')
    clb3 = fig.colorbar(img3, cax=cax3)
    clb3.set_label(r'$\frac{I}{I_{phot}}$', fontsize=plotp['fontsize'])

    ax1.set_xticks(np.arange(0, 420, 60))
    ax1.set_yticks(np.arange(-90, 120, 30))
    ax1.set_title('Artificial Map', fontsize=plotp['fontsize'])
    ax1.set_xlabel(r'Longitude ($^\circ$)', fontsize=plotp['fontsize'])
    ax1.set_ylabel(r'Latitude ($^\circ$)', fontsize=plotp['fontsize'])

    ax2.set_xticks(np.arange(0, 420, 60))
    ax2.set_yticks(np.arange(-90, 120, 30))
    ax2.set_title('Reconstructed Map', fontsize=plotp['fontsize'])
    ax2.set_xlabel(r'Longitude ($^\circ$)', fontsize=plotp['fontsize'])
    ax2.set_ylabel(r'Latitude ($^\circ$)', fontsize=plotp['fontsize'])

    colors = ['r', 'b', 'g', 'purple']
    for i, mode in enumerate(DIP.conf):
        epochs = (DIP.idc[mode]['times'] - DIP.params['t0']) / DIP.params['period']
        phases = epochs - np.floor(epochs)

        ax1.plot([360 * (1.0 - phases), 360 * (1.0 - phases)], [-85, -75], '-', color=colors[i], linewidth=2)
        ax2.plot([360 * (1.0 - phases), 360 * (1.0 - phases)], [-85, -75], '-', color=colors[i], linewidth=2)

        if 0.0 in phases:
            ax1.plot([0, 0], [-85, -75], '-', color=colors[i], linewidth=2)
            ax2.plot([0, 0], [-85, -75], '-', color=colors[i], linewidth=2)

    ax1.tick_params(axis='both', labelsize=plotp['ticklabelsize'])
    ax2.tick_params(axis='both', labelsize=plotp['ticklabelsize'])

    plt.tight_layout()


def lmbds_plot(DIP, plotp):

    lmbds = DIP.opt_results['lmbds']
    chisqs = DIP.opt_results['total_chisqs']
    mems = DIP.opt_results['mems']
    maxcurve = DIP.opt_results['maxcurve']

    fig, ax1 = plt.subplots()

    lmbd = lmbds[maxcurve]

    ax1.plot(chisqs, mems, 'ko', ms=2)
    ax1.plot(chisqs[maxcurve], mems[maxcurve], 'ro', ms=3, label='Best lmbd = ' + str(lmbd))
    ax1.set_xlabel(r'$\chi^2$', fontsize=plotp['fontsize'])
    ax1.set_ylabel('Entropy', fontsize=plotp['fontsize'])
    ax1.tick_params(axis='both', labelsize=plotp['ticklabelsize'])
    ax1.legend()

    plt.tight_layout()


def make_grid_contours(chisq_grid):

    names = list(chisq_grid.keys())
    names.remove('chisqs')

    mn = np.argmin(chisq_grid['chisqs'])

    if len(names) == 1:

        mpar = chisq_grid[names[0]][mn]

        print(names[0], ':', mpar)

        plt.plot(chisq_grid[names[0]], chisq_grid['chisqs'], 'ko')
        plt.xlabel(names[0], fontsize=15)
        if names[0] == 'vsini':
            plt.xlabel('vsini (km/s)', fontsize=15)

        elif names[0] == 'eqw':
            plt.xlabel('EW', fontsize=15)
        plt.ylabel(r'$\chi^2$', fontsize=15)
        plt.show()

        return

    combin = np.array(list(combinations(names, 2)))
    combin[len(names) - 2] = combin[len(names) - 2][::-1]

    print('\n' + '\033[96m' + '*** Grid Search Results ***' + '\033[0m')
    for pair in combin:

        par1_name = pair[0]
        par2_name = pair[1]

        index = []
        for rpar in chisq_grid:
            if rpar not in [par1_name, par2_name, 'chisqs']:
                index.append(np.argwhere(chisq_grid[rpar] == chisq_grid[rpar][mn]).flatten())

        if len(index) > 0:
            inters = reduce(np.intersect1d, index)
            chi = np.array(chisq_grid['chisqs'])[inters]
            par1 = chisq_grid[par1_name][inters]
            par2 = chisq_grid[par2_name][inters]
        else:
            chi = np.array(chisq_grid['chisqs'])
            par1 = chisq_grid[par1_name]
            par2 = chisq_grid[par2_name]

        contour = np.zeros((len(np.unique(par2)), len(np.unique(par1))))
        for i, ei in enumerate(sorted(np.unique(par2))):
            vss = np.argsort(par1[par2 == ei])
            contour[i] = chi[par2 == ei][vss]

        mn = np.argmin(chisq_grid['chisqs'])
        mpar1 = chisq_grid[par1_name][mn]
        mpar2 = chisq_grid[par2_name][mn]

        fig = plt.figure()
        levels = [1 + chisq_grid['chisqs'][mn]]
        uc = plt.contour(sorted(np.unique(par1)), sorted(np.unique(par2)), contour, levels=levels,
                         cmap='RdYlBu')  # alpha=0.7,
        plt.show(block=False)
        plt.close(fig=fig)

        chipone = uc.allsegs[0][0]

        if len(chipone) > 0:
            print('\033[93m' + '{:<25}'.format(par1_name) + ':' + '\033[0m',
                  '\033[1m' + str(mpar1) + '\033[0m', '\033[1m' + '-' + str(mpar1 - min(chipone[:, 0])) + '\033[0m',
                  '\033[1m' + '+' + str(max(chipone[:, 0]) - mpar1) + '\033[0m')
            print('\033[93m' + '{:<25}'.format(par2_name) + ':' + '\033[0m',
                  '\033[1m' + str(mpar2) + '\033[0m', '\033[1m' + '-' + str(mpar2 - min(chipone[:, 1])) + '\033[0m',
                  '\033[1m' + '+' + str(max(chipone[:, 1]) - mpar2) + '\033[0m')
        else:
            print('\033[93m' + '{:<25}'.format(par1_name) + ':' + '\033[0m', '\033[1m' + str(mpar1) + '\033[0m')
            print('\033[93m' + '{:<25}'.format(par2_name) + ':' + '\033[0m', '\033[1m' + str(mpar2) + '\033[0m')

        fig = plt.figure()
        plt.contourf(sorted(np.unique(par1)), sorted(np.unique(par2)), contour, 50, cmap='RdYlBu')  # alpha=0.7,
        plt.xlabel(par1_name, fontsize=15)
        plt.ylabel(par2_name, fontsize=15)

        plt.tick_params(labelsize=15)

        cbaxes = fig.add_axes((0.150, 0.85, 0.82, 0.03))
        cbar = plt.colorbar(cax=cbaxes, orientation='horizontal',
                            ticks=np.linspace(np.min(contour), np.max(contour), 5), format="%0.3f")
        cbar.ax.tick_params(labelsize=15)
        cbar.set_label('$\chi^2$', fontsize=15)
        cbar.ax.xaxis.set_ticks_position("top")
        cbar.ax.xaxis.set_label_position("top")

        plt.subplots_adjust(left=0.15, bottom=0.14, right=0.97, top=0.84)

    plt.show()


def grid_test(xyzs, scalars, title, show=True):

    from mayavi import mlab

    def plot(dx, dy, dz, dtriangles, dnscalars1):
        mlab.figure(title)
        smesh = mlab.triangular_mesh(dx, dy, dz, dtriangles, scalars=dnscalars1,
                                     colormap='YlOrBr', line_width=3.0)

        carr = smesh.module_manager.scalar_lut_manager.lut.table.to_array()
        ncarr = carr[::-1]
        smesh.module_manager.scalar_lut_manager.lut.table = ncarr

        cb = mlab.colorbar(smesh, title=title, nb_labels=5, label_fmt='%0.4f', orientation='horizontal')
        cb.label_text_property.font_family = 'times'
        cb.label_text_property.bold = 0
        # cb.label_text_property.font_size = 50

    x = []
    y = []
    z = []
    triangles = []
    nscalars = []
    for i, xyz in enumerate(xyzs):
        nscalars.append([scalars[i]] * 3)
        for row in xyz:
            x.append(row[0])
            y.append(row[1])
            z.append(row[2])
        triangles.append((0 + i * 3, 1 + i * 3, 2 + i * 3))
    nscalars = np.hstack(nscalars)

    plot(x, y, z, triangles, nscalars)

    if show:
        mlab.show()


def draw_3D_surf(params, surface_grid, spots_params):

    fssc, fssh = generate_spotted_surface(surface_grid, spots_params)

    Thot = params['Thot']
    Tphot = params['Tphot']
    Tcool = params['Tcool']
    xyzs = surface_grid['grid_xyzs']

    brightness = (fssc * Tcool ** 4 + fssh * Thot ** 4 + (1.0 - (fssc + fssh)) * Tphot ** 4) / Tphot ** 4

    if surface_grid['method'] != 'phoebe2_marching':
        scalars = np.repeat(brightness, 2)
    else:
        scalars = brightness

    grid_test(xyzs=xyzs, scalars=scalars, title='Spotted Star')


def mp_calc_pixel_coeffs(cpu_num, input_args, mode):

    if mode != 'lc':
        func = cutils.calc_pixel_coeffs_spec
    else:
        func = cutils.calc_pixel_coeffs_lc

    if cpu_num > 1:
        pool = multiprocessing.Pool(processes=cpu_num)
        results = pool.map(func, input_args)
        pool.close()
        pool.join()

    else:
        results = []
        for item in input_args:
            results.append(func(item))

    return results


def mp_search(cpu_num, func, input_args):

    pool = multiprocessing.Pool(processes=cpu_num)
    results = []
    for result in tqdm.tqdm(pool.imap(func=func, iterable=input_args), total=len(input_args)):
        results.append(result)

    return results


def isfloatable(x):

    if x is None or isinstance(x, (bool, np.bool_)):
        return False

    try:
        float(x)
        return True

    except ValueError:
        return False



# cython: language_level=3

cimport cython

import numpy as np
cimport numpy as np
# from cython cimport Py_ssize_t
from libc.math cimport sin, cos, atan2, log10, sqrt, pi, exp
from libc.stdlib cimport malloc, free
# from cython.parallel import prange

DTYPE = np.float64
ctypedef np.float64_t DTYPE_t

cdef DTYPE_t solMass = 1.988409870698051e+30  # kg
cdef DTYPE_t solRad = 695700000.0             # m
cdef DTYPE_t G = 6.6743e-11                   # m3 / (kg s2)

cpdef tuple spherical_to_cartesian(DTYPE_t r, DTYPE_t lat, DTYPE_t lon):
    cdef DTYPE_t x, y, z
    x = r * cos(lat) * cos(lon)
    y = r * cos(lat) * sin(lon)
    z = r * sin(lat)
    return x, y, z

cpdef tuple cartesian_to_spherical(DTYPE_t x, DTYPE_t y, DTYPE_t z):
    cdef DTYPE_t r, lat, lon
    r = sqrt(x**2 + y**2 + z**2)
    lat = atan2(z, sqrt(x**2 + y**2))
    lon = atan2(y, x)
    lon = lon + 2.0 * pi if lon < 0 else lon

    return r, lat, lon

cpdef tuple cartesian_to_spherical_vectorized(np.ndarray[DTYPE_t, ndim=1] x,
                                              np.ndarray[DTYPE_t, ndim=1] y,
                                              np.ndarray[DTYPE_t, ndim=1] z):
    cdef np.ndarray[DTYPE_t, ndim=1] r = np.sqrt(x**2 + y**2 + z**2)
    cdef np.ndarray[DTYPE_t, ndim=1] lat = np.arctan2(z, np.sqrt(x**2 + y**2))
    cdef np.ndarray[DTYPE_t, ndim=1] lon = np.arctan2(y, x)
    lon = np.where(lon < 0, lon + 2.0 * pi, lon)  # Negatif uzunlukları 0-2π aralığına getir

    return r, lat, lon

cpdef DTYPE_t newtonRaphson(DTYPE_t x0, DTYPE_t omega, DTYPE_t theta, DTYPE_t tol=1.0e-8):
    cdef DTYPE_t h

    h = r_func(x0, omega, theta) / deriv_r_func(x0, omega, theta)
    while abs(h) >= tol:
        h = r_func(x0, omega, theta) / deriv_r_func(x0, omega, theta)
        x0 = x0 - h

    return x0

cpdef DTYPE_t r_func(DTYPE_t r, DTYPE_t omega, DTYPE_t theta):

    return (1.0 / omega ** 2) + 0.5 - (1.0 / (omega ** 2 * r)) - 0.5 * r ** 2 * sin(theta)**2

cpdef DTYPE_t deriv_r_func(DTYPE_t r, DTYPE_t omega, DTYPE_t theta):

    return 1.0 / omega**2 / r**2 - sin(theta)**2 * r

cpdef DTYPE_t calc_r_to_Re_ratio(DTYPE_t omega, DTYPE_t theta):
    cdef DTYPE_t x0=1.0

    return newtonRaphson(x0=x0, omega=omega, theta=theta)

cpdef DTYPE_t calc_triangle_area(np.ndarray[DTYPE_t, ndim=2] triangle):

    cdef DTYPE_t area
    cdef np.ndarray[DTYPE_t, ndim=1] cp = np.cross(triangle[1] - triangle[0], triangle[2] - triangle[0])

    area = np.linalg.norm(cp) / 2.0

    return area

cpdef DTYPE_t calc_logg(DTYPE_t mass, DTYPE_t re, DTYPE_t rs, DTYPE_t omega, DTYPE_t theta):
    cdef DTYPE_t mass_kg = mass * solMass
    cdef DTYPE_t g_cm = mass_kg * G * 1.0e6
    cdef DTYPE_t r_cm = re * solRad * 100
    cdef DTYPE_t r_ = rs / re
    cdef DTYPE_t geff

    theta = pi - theta if theta > pi / 2. else theta

    geff = (g_cm / r_cm**2) * sqrt(1.0 / r_**4 + omega**4 * r_**2 * sin(theta)**2 - 2.0 * omega**2 * sin(theta)**2 / r_)
    return log10(geff)

cpdef DTYPE_t ld_factors_calc(str ld_law, DTYPE_t mu, DTYPE_t ldc1, DTYPE_t ldc2):

    if ld_law == 'linear':
        return 1. - ldc1 * (1. - mu)

    elif ld_law == 'square-root':
        return 1. - ldc1 * (1. - mu) - ldc2 * (1. - sqrt(mu))

    elif ld_law == 'quadratic':
        return 1. - ldc1 * (1. - mu) - ldc2 * (1. - mu) ** 2

cpdef tuple td_grid_sect(tuple args):

    cdef DTYPE_t[2] blatpn = args[0]
    cdef DTYPE_t uarea = args[1]
    cdef DTYPE_t omega = args[2]
    cdef DTYPE_t radius = args[3]
    cdef DTYPE_t mass = args[4]
    cdef int nlon = int(round(2 * pi * radius ** 2 * abs(sin(blatpn[0]) - sin(blatpn[1])) / uarea))
    cdef np.ndarray[DTYPE_t, ndim=1] blong = np.linspace(0.0, 2.0 * pi, nlon + 1, dtype=DTYPE)
    cdef list areas = []
    cdef list xyzs = []
    cdef list xpc = []
    cdef list ypc = []
    cdef list zpc = []
    cdef list rs = []
    cdef list lats = []
    cdef list longs = []
    cdef list loggs = []
    cdef int i, j, k
    cdef list inds = [[(0, 0), (1, 0), (0, 1)], [(1, 0), (1, 1), (0, 1)]]
    cdef DTYPE_t rzlat, rlat, xp, yp, zp, x, y, z

    rzlat = calc_r_to_Re_ratio(omega=omega, theta=(blatpn[0] + blatpn[1] + pi) / 2.0) * radius

    for j in range(nlon):
        for i in range(2):
            polygon = []
            for k in range(3):
                rlat = calc_r_to_Re_ratio(omega=omega, theta=blatpn[inds[i][k][0]] + pi / 2.0) * radius
                xp, yp, zp = spherical_to_cartesian(rlat, blatpn[inds[i][k][0]], blong[j + inds[i][k][1]])
                polygon.append([xp, yp, zp])
            xyzs.append(np.array(polygon))
            areas.append(calc_triangle_area(np.array(polygon)))

        rs.append(rzlat)
        lats.append((blatpn[0] + blatpn[1]) / 2.0)
        longs.append((blong[j] + blong[j + 1]) / 2.0)

        x, y, z = spherical_to_cartesian(rzlat, (blatpn[0] + blatpn[1]) / 2.0, (blong[j] + blong[j + 1]) / 2.0)

        xpc.append(x)
        ypc.append(y)
        zpc.append(z)

        loggs.append(calc_logg(mass=mass, re=radius, rs=rzlat, omega=omega, theta=((blatpn[0] + blatpn[1])
                                                                                   / 2.0) + pi / 2.0))

    return nlon, blong, areas, rs, lats, longs, xyzs, loggs

cpdef list td_surface_grid(np.ndarray[DTYPE_t, ndim=1] blats, DTYPE_t omega, DTYPE_t radius, DTYPE_t mass):

    cdef int j
    cdef int ipn = 4
    cdef DTYPE_t uarea, pzh
    cdef list results = []

    pzh = abs(sin(blats[0]) - sin(blats[1])) * radius
    uarea = 2. * pi * radius * pzh / ipn

    for j in range(len(blats) - 1):
        results.append(td_grid_sect(([blats[j], blats[j + 1]], uarea, omega, radius, mass)))

    # parts = np.array(results, dtype=object)
    #
    # container['method'] = 'trapezoid'
    # container['nlons'] = parts[:, 0]
    # container['blongs'] = parts[:, 1]
    # container['grid_areas'] = np.hstack(parts[:, 2]).reshape(-1, 2).sum(axis=1)
    # container['grid_rs'] = np.hstack(parts[:, 3])
    # container['grid_lats'] = np.hstack(parts[:, 4])
    # container['grid_longs'] = np.hstack(parts[:, 5])
    # container['grid_xyzs'] = np.vstack(parts[:, 6])
    # container['grid_loggs'] = np.hstack(parts[:, 7])
    # container['blats'] = blats

    return results

cpdef tuple macro_kernel(int nop, DTYPE_t muval, DTYPE_t vrt, DTYPE_t deltav):

    """ Adapted from SME (Valenti, J. A., & Piskunov, N. 1996, A&AS, 118, 595) """

    cdef DTYPE_t sigma, sigr, sigt, nfine
    cdef int nmk = 3

    sigma = vrt / sqrt(2.0) / deltav
    sigr = sigma * muval
    sigt = sigma * sqrt(1.0 - muval ** 2)
    nfine = nop

    cdef DTYPE_t temp1, temp2, mrsum = 0.0, mtsum = 0.0

    temp1 = 10.0 * sigma
    if temp1 > nmk:
        nmk = <int> temp1

    temp2 = (nfine - 3) / 2
    if temp2 > nmk:
        nmk = <int> temp2

    if nmk < 3:
        nmk = 3

    cdef np.ndarray[DTYPE_t, ndim=1] mrkern = np.zeros(2 * nmk + 1, dtype=DTYPE)
    cdef np.ndarray[DTYPE_t, ndim=1] mtkern = np.zeros(2 * nmk + 1, dtype=DTYPE)
    cdef int i, j

    if sigr > 0.0:
        for i in range(2 * nmk + 1):
            if -0.5 * ((i - nmk) / sigr)**2 < -20.0:
                mrkern[i] = exp(-20)
            else:
                mrkern[i] = exp(-0.5 * ((i - nmk) / sigr)**2)
            mrsum = mrsum + mrkern[i]
        mrkern /= mrsum
        # mrkern /= np.sum(mrkern)

    else:
        mrkern[nmk] = 1.0

    if sigt > 0.0:
        for j in range(2 * nmk + 1):
            if -0.5 * ((j - nmk) / sigt)**2 < -20.0:
                mtkern[j] = exp(-20)
            else:
                mtkern[j] = exp(-0.5 * ((j - nmk) / sigt)**2)
            mtsum += mtkern[j]
        mtkern /= mtsum
        # mtkern /= np.sum(mtkern)

    else:
        mtkern[nmk] = 1.0

    return (mrkern + mtkern) / 2.0, nmk

cpdef apply_macro_conv(np.ndarray[DTYPE_t, ndim=1] vels, np.ndarray[DTYPE_t, ndim=1] ints, DTYPE_t mu, DTYPE_t vrt):

    cdef DTYPE_t deltav = vels[1] - vels[0]

    mkernel, nmk = macro_kernel(len(vels), mu, vrt, deltav)

    conv_mac = np.convolve(np.pad(ints, (nmk, nmk), 'edge'), mkernel, mode='same')

    return conv_mac, nmk

# @cython.boundscheck(False)
# @cython.wraparound(False)
cpdef linear_interpolation(DTYPE_t[:] x, DTYPE_t[:] y, DTYPE_t[:] x_interp):
    cdef int n = len(x)
    cdef int m = len(x_interp)
    cdef DTYPE_t[:] y_interp = np.zeros(m, dtype=DTYPE)

    cdef int i, j
    cdef DTYPE_t slope, intercept

    for j in range(m):
        i = 0
        while i < n and x_interp[j] > x[i]:
            i += 1

        if i == 0:
            slope = (y[1] - y[0]) / (x[1] - x[0])
            intercept = y[0] - slope * x[0]
        elif i == n:
            slope = (y[n - 1] - y[n - 2]) / (x[n - 1] - x[n - 2])
            intercept = y[n - 1] - slope * x[n - 1]
        else:
            slope = (y[i] - y[i - 1]) / (x[i] - x[i - 1])
            intercept = y[i] - slope * x[i]

        y_interp[j] = slope * x_interp[j] + intercept

    return y_interp

# @cython.boundscheck(False)
# @cython.wraparound(False)
cpdef np.ndarray[DTYPE_t, ndim=2] calc_pixel_coeffs_spec(tuple args):

    cdef DTYPE_t itime = args[0]
    cdef np.ndarray[DTYPE_t, ndim=1] plats = args[1]
    cdef np.ndarray[DTYPE_t, ndim=1] vlats = args[2]
    cdef np.ndarray[DTYPE_t, ndim=2] ldcs_phot = args[3]
    cdef np.ndarray[DTYPE_t, ndim=2] ldcs_cool = args[4]
    cdef np.ndarray[DTYPE_t, ndim=2] ldcs_hot = args[5]
    cdef np.ndarray[DTYPE_t, ndim=1] lis_phot = args[6]
    cdef np.ndarray[DTYPE_t, ndim=1] lis_cool = args[7]
    cdef np.ndarray[DTYPE_t, ndim=1] lis_hot = args[8]
    cdef np.ndarray[DTYPE_t, ndim=1] areas = args[9]
    cdef np.ndarray[DTYPE_t, ndim=1] grid_lats = args[10]
    cdef np.ndarray[DTYPE_t, ndim=1] grid_longs = args[11]
    cdef int nop = args[12]
    cdef DTYPE_t t0 = args[13]
    cdef DTYPE_t incl = args[14]
    cdef str ld_law = args[15]
    cdef int noes = args[16]
    cdef np.ndarray[DTYPE_t, ndim=1] lp_vels = args[17]
    cdef np.ndarray[DTYPE_t, ndim=1] phot_lp_data = args[18]
    cdef np.ndarray[DTYPE_t, ndim=1] cool_lp_data = args[19]
    cdef np.ndarray[DTYPE_t, ndim=1] hot_lp_data = args[20]
    cdef DTYPE_t vrt = args[21]
    cdef np.ndarray[DTYPE_t, ndim=1] vels = args[22]
    cdef DTYPE_t period = args[23]
    cdef int info = args[24]

    cdef np.ndarray[DTYPE_t, ndim=2] coeffs_cube = np.zeros((noes, 3 + nop * 3), dtype=np.float64)
    cdef int i
    cdef DTYPE_t nlong, iepoch
    cdef DTYPE_t ldf_phot, ldf_cool, ldf_hot, mu, cosincl = cos(incl), sinincl = sin(incl)

    if info:
        iepoch = (itime - t0) / period
        print('\033[94m  mid-time = {0:.5f} - mid-epoch = {1:.5f} \033[0m'.format(itime, iepoch))

    for i in range(noes):

        nlong = grid_longs[i] + 2.0 * pi * (itime - t0) / plats[i]
        mu = sin(grid_lats[i]) * cosincl + cos(grid_lats[i]) * sinincl * cos(nlong)

        if mu > 0.0:

            ldf_phot = ld_factors_calc(ld_law, mu, ldcs_phot[i, 0], ldcs_phot[i, 1])
            ldf_cool = ld_factors_calc(ld_law, mu, ldcs_cool[i, 0], ldcs_cool[i, 1])
            ldf_hot = ld_factors_calc(ld_law, mu, ldcs_hot[i, 0], ldcs_hot[i, 1])

            coeffs_cube[i, 0] = lis_phot[i] * ldf_phot * areas[i] * mu
            coeffs_cube[i, 1] = lis_cool[i] * ldf_cool * areas[i] * mu
            coeffs_cube[i, 2] = lis_hot[i] * ldf_hot * areas[i] * mu


            dv = vlats[i] * sin(nlong) * sinincl

            pcm_phot, pnmk = apply_macro_conv(lp_vels, phot_lp_data, mu, vrt)
            pcm_cool, cnmk = apply_macro_conv(lp_vels, cool_lp_data, mu, vrt)
            pcm_hot, hnmk = apply_macro_conv(lp_vels, hot_lp_data, mu, vrt)

            lp_phot = np.interp(vels, lp_vels + dv, pcm_phot[pnmk: -pnmk])
            lp_cool = np.interp(vels, lp_vels + dv, pcm_cool[cnmk: -cnmk])
            lp_hot = np.interp(vels, lp_vels + dv, pcm_hot[hnmk: -hnmk])

            # lp_phot = linear_interpolation(lp_vels + dv, pcm_phot[pnmk: -pnmk], vels)
            # lp_cool = linear_interpolation(lp_vels + dv, pcm_cool[cnmk: -cnmk], vels)
            # lp_hot = linear_interpolation(lp_vels + dv, pcm_hot[hnmk: -hnmk], vels)

            coeffs_cube[i, 3: 3 + nop] = lp_phot
            coeffs_cube[i, 3 + nop: 3 + 2 * nop] = lp_cool
            coeffs_cube[i, 3 + 2 * nop:] = lp_hot

    return coeffs_cube

cpdef np.ndarray[DTYPE_t, ndim=2] calc_pixel_coeffs_lc(tuple args):

    cdef DTYPE_t itime = args[0]
    cdef np.ndarray[DTYPE_t, ndim=1] plats = args[1]
    cdef np.ndarray[DTYPE_t, ndim=2] ldcs_phot = args[2]
    cdef np.ndarray[DTYPE_t, ndim=2] ldcs_cool = args[3]
    cdef np.ndarray[DTYPE_t, ndim=2] ldcs_hot = args[4]
    cdef np.ndarray[DTYPE_t, ndim=1] lis_phot = args[5]
    cdef np.ndarray[DTYPE_t, ndim=1] lis_cool = args[6]
    cdef np.ndarray[DTYPE_t, ndim=1] lis_hot = args[7]
    cdef np.ndarray[DTYPE_t, ndim=1] areas = args[8]
    cdef np.ndarray[DTYPE_t, ndim=1] grid_lats = args[9]
    cdef np.ndarray[DTYPE_t, ndim=1] grid_longs = args[10]
    cdef DTYPE_t t0 = args[11]
    cdef DTYPE_t incl = args[12]
    cdef str ld_law = args[13]
    cdef int noes = args[14]
    cdef DTYPE_t period = args[15]
    cdef int info = args[16]

    cdef np.ndarray[DTYPE_t, ndim=2] coeffs_cube = np.zeros((noes, 3), dtype=np.float64)
    cdef int i
    cdef DTYPE_t nlong
    cdef DTYPE_t ldf_phot, ldf_cool, ldf_hot, mu, cosincl = cos(incl), sinincl = sin(incl)

    if info:
        iepoch = (itime - t0) / period
        print('\033[94m  mid-time = {0:.5f} - mid-epoch = {1:.5f} \033[0m'.format(itime, iepoch))

    for i in range(noes):

        nlong = grid_longs[i] + 2.0 * pi * (itime - t0) / plats[i]
        mu = sin(grid_lats[i]) * cosincl + cos(grid_lats[i]) * sinincl * cos(nlong)

        if mu > 0.0:

            ldf_phot = ld_factors_calc(ld_law, mu, ldcs_phot[i, 0], ldcs_phot[i, 1])
            ldf_cool = ld_factors_calc(ld_law, mu, ldcs_cool[i, 0], ldcs_cool[i, 1])
            ldf_hot = ld_factors_calc(ld_law, mu, ldcs_hot[i, 0], ldcs_hot[i, 1])

            coeffs_cube[i, 0] = lis_phot[i] * ldf_phot * areas[i] * mu
            coeffs_cube[i, 1] = lis_cool[i] * ldf_cool * areas[i] * mu
            coeffs_cube[i, 2] = lis_hot[i] * ldf_hot * areas[i] * mu

    return coeffs_cube

# @cython.boundscheck(False)
# @cython.wraparound(False)
# def subdivision(np.ndarray[DTYPE_t, ndim=3] xyzs, np.ndarray[DTYPE_t, ndim=1] fss, int subno=1):
#     cdef np.ndarray[DTYPE_t, ndim=3] xyzsb = xyzs.copy()
#     cdef np.ndarray[DTYPE_t, ndim=1] fssb = fss.copy()
#     cdef int i, j, k, n, m
#     cdef double x1, y1, z1, x2, y2, z2, x3, y3, z3
#     cdef double xa, ya, za, xb, yb, zb, xc, yc, zc
#     cdef np.ndarray[DTYPE_t, ndim=2] tri
#     cdef np.ndarray[DTYPE_t, ndim=2] new_tri
#     cdef np.ndarray[DTYPE_t, ndim=1] xcs, ycs, zcs
#     cdef list tri_all, rd_fss
#
#     for i in range(subno):
#         rd_fss = []
#         tri_all = []
#         for j in range(len(xyzsb)):
#             x1, y1, z1 = xyzsb[j, 0, 0], xyzsb[j, 0, 1], xyzsb[j, 0, 2]
#             x2, y2, z2 = xyzsb[j, 1, 0], xyzsb[j, 1, 1], xyzsb[j, 1, 2]
#             x3, y3, z3 = xyzsb[j, 2, 0], xyzsb[j, 2, 1], xyzsb[j, 2, 2]
#
#             xa, ya, za = (x1 + x2) / 2., (y1 + y2) / 2., (z1 + z2) / 2.
#             xb, yb, zb = (x1 + x3) / 2., (y1 + y3) / 2., (z1 + z3) / 2.
#             xc, yc, zc = (x2 + x3) / 2., (y2 + y3) / 2., (z2 + z3) / 2.
#
#             tri_all.append(np.array([[x1, y1, z1], [xa, ya, za], [xb, yb, zb]], dtype=np.float64))
#             tri_all.append(np.array([[xa, ya, za], [x2, y2, z2], [xc, yc, zc]], dtype=np.float64))
#             tri_all.append(np.array([[xc, yc, zc], [x3, y3, z3], [xb, yb, zb]], dtype=np.float64))
#             tri_all.append(np.array([[xa, ya, za], [xb, yb, zb], [xc, yc, zc]], dtype=np.float64))
#
#             rd_fss.append([fssb[j]] * 4)
#
#         fssb = np.hstack(rd_fss)
#         xyzsb = np.array(tri_all, dtype=np.float64)
#
#     n = len(xyzsb)
#     rs = np.empty(n, dtype=np.float64)
#     lats = np.empty(n, dtype=np.float64)
#     longs = np.empty(n, dtype=np.float64)
#
#     for j in range(n):
#         tri = xyzsb[j]
#         xc = (tri[0, 0] + tri[1, 0] + tri[2, 0]) / 3.
#         yc = (tri[0, 1] + tri[1, 1] + tri[2, 1]) / 3.
#         zc = (tri[0, 2] + tri[1, 2] + tri[2, 2]) / 3.
#
#         r, lat, long = cartesian_to_spherical(xc, yc, zc)
#
#         rs[j] = r
#         lats[j] = lat
#         longs[j] = long
#
#     return xyzsb, fssb, r, lats, longs

def subdivision(np.ndarray[DTYPE_t, ndim=3] xyzs, np.ndarray[DTYPE_t, ndim=1] fss, int subno=1):
    cdef np.ndarray[DTYPE_t, ndim=3] xyzsb = xyzs.copy()
    cdef np.ndarray[DTYPE_t, ndim=1] fssb = fss.copy()
    cdef int i, j, n
    cdef double[:, ::1] midpoints
    cdef np.ndarray[DTYPE_t, ndim=2] tri
    cdef np.ndarray[DTYPE_t, ndim=3] new_xyzsb
    cdef np.ndarray[DTYPE_t, ndim=1] new_fssb

    for i in range(subno):
        n = len(xyzsb) * 4
        new_xyzsb = np.empty((n, 3, 3), dtype=np.float64)
        new_fssb = np.repeat(fssb, 4)

        for j in range(len(xyzsb)):
            tri = xyzsb[j]
            x1, y1, z1 = tri[0]
            x2, y2, z2 = tri[1]
            x3, y3, z3 = tri[2]

            xa, ya, za = (x1 + x2) / 2., (y1 + y2) / 2., (z1 + z2) / 2.
            xb, yb, zb = (x1 + x3) / 2., (y1 + y3) / 2., (z1 + z3) / 2.
            xc, yc, zc = (x2 + x3) / 2., (y2 + y3) / 2., (z2 + z3) / 2.

            new_xyzsb[j * 4 + 0] = [[x1, y1, z1], [xa, ya, za], [xb, yb, zb]]
            new_xyzsb[j * 4 + 1] = [[xa, ya, za], [x2, y2, z2], [xc, yc, zc]]
            new_xyzsb[j * 4 + 2] = [[xc, yc, zc], [x3, y3, z3], [xb, yb, zb]]
            new_xyzsb[j * 4 + 3] = [[xa, ya, za], [xb, yb, zb], [xc, yc, zc]]

        xyzsb = new_xyzsb
        fssb = new_fssb

    n = len(xyzsb)
    rs = np.empty(n, dtype=np.float64)
    lats = np.empty(n, dtype=np.float64)
    longs = np.empty(n, dtype=np.float64)

    # Tüm üçgenler için ortalama koordinatları hesapla
    xcs = (xyzsb[:, 0, 0] + xyzsb[:, 1, 0] + xyzsb[:, 2, 0]) / 3
    ycs = (xyzsb[:, 0, 1] + xyzsb[:, 1, 1] + xyzsb[:, 2, 1]) / 3
    zcs = (xyzsb[:, 0, 2] + xyzsb[:, 1, 2] + xyzsb[:, 2, 2]) / 3

    # Kartesyen'den küresel koordinatlara dönüştürme
    rs, lats, longs = cartesian_to_spherical_vectorized(xcs, ycs, zcs)

    return xyzsb, fssb, rs, lats, longs


# @cython.boundscheck(False)
# @cython.wraparound(False)
# def subdivision(np.ndarray[DTYPE_t, ndim=3] xyzs, np.ndarray[DTYPE_t, ndim=1] fss, int subno=1):
#     cdef np.ndarray[DTYPE_t, ndim=3] xyzsb = xyzs.copy()
#     cdef np.ndarray[DTYPE_t, ndim=1] fssb = fss.copy()
#     cdef int i, j
#     cdef double x1, y1, z1, x2, y2, z2, x3, y3, z3
#     cdef double xa, ya, za, xb, yb, zb, xc, yc, zc
#     cdef list tri_all, rd_fss
#     cdef np.ndarray[DTYPE_t, ndim=2] tri
#     cdef np.ndarray[DTYPE_t, ndim=2] new_tri
#     cdef np.ndarray[DTYPE_t, ndim=1] xcs, ycs, zcs
#     cdef double xc, yc, zc
#
#     for i in range(subno):
#         rd_fss = []
#         tri_all = []
#         for j in range(len(xyzsb)):
#             x1, y1, z1 = xyzsb[j, 0, 0], xyzsb[j, 0, 1], xyzsb[j, 0, 2]
#             x2, y2, z2 = xyzsb[j, 1, 0], xyzsb[j, 1, 1], xyzsb[j, 1, 2]
#             x3, y3, z3 = xyzsb[j, 2, 0], xyzsb[j, 2, 1], xyzsb[j, 2, 2]
#
#             xa, ya, za = (x1 + x2) / 2., (y1 + y2) / 2., (z1 + z2) / 2.
#             xb, yb, zb = (x1 + x3) / 2., (y1 + y3) / 2., (z1 + z3) / 2.
#             xc, yc, zc = (x2 + x3) / 2., (y2 + y3) / 2., (z2 + z3) / 2.
#
#             tri_all.append(np.array([[x1, y1, z1], [xa, ya, za], [xb, yb, zb]], dtype=DTYPE_t))
#             tri_all.append(np.array([[xa, ya, za], [x2, y2, z2], [xc, yc, zc]], dtype=DTYPE_t))
#             tri_all.append(np.array([[xc, yc, zc], [x3, y3, z3], [xb, yb, zb]], dtype=DTYPE_t))
#             tri_all.append(np.array([[xa, ya, za], [xb, yb, zb], [xc, yc, zc]], dtype=DTYPE_t))
#
#             rd_fss.append([fssb[j]] * 4)
#
#         fssb = np.hstack(rd_fss)
#         xyzsb = np.array(tri_all, dtype=DTYPE_t)
#
#     xcs = np.empty(len(xyzsb), dtype=DTYPE_t)
#     ycs = np.empty(len(xyzsb), dtype=DTYPE_t)
#     zcs = np.empty(len(xyzsb), dtype=DTYPE_t)
#
#     for j in range(len(xyzsb)):
#         tri = xyzsb[j]
#         xc = (tri[0, 0] + tri[1, 0] + tri[2, 0]) / 3.
#         yc = (tri[0, 1] + tri[1, 1] + tri[2, 1]) / 3.
#         zc = (tri[0, 2] + tri[1, 2] + tri[2, 2]) / 3.
#
#         xcs[j] = xc
#         ycs[j] = yc
#         zcs[j] = zc
#
#
#     return , fssb

# cpdef tuple macro_kernel(int nop, double muval, double vrt, double deltav):
#
#     cdef int os
#     cdef double area_r = 0.5
#     cdef double area_t = 0.5
#     cdef double sigma, sigr, sigt, nfine
#     cdef int nmk
#
#     cdef np.ndarray[np.double_t, ndim=1] xarg_r, exparg_r, xarg_t, exparg_t, mrkern, mtkern, mkern
#
#     # cdef double[:] xarg_r, exparg_r, xarg_t, exparg_t, mrkern, mtkern, mkern
#
#     os = 1
#     sigma = os * vrt / np.sqrt(2.0) / deltav
#     sigr = sigma * muval
#     sigt = sigma * np.sqrt(1.0 - muval ** 2)
#     nfine = os * nop
#
#     nmk = max(int(10.0 * sigma), int((nfine - 3) / 2), 3)
#
#     if sigr > 0.0:
#         xarg_r = (np.arange(2 * nmk + 1, dtype=np.int) - nmk) / sigr
#         exparg_r = (-0.5 * np.power(xarg_r, 2))
#         exparg_r = np.where(exparg_r < -20.0, -20.0, exparg_r)
#         mrkern = np.exp(exparg_r)
#         mrkern = mrkern / np.sum(mrkern)
#     else:
#         mrkern = np.zeros(2 * nmk + 1)
#         mrkern[nmk] = 1.0
#
#     if sigt > 0.0:
#         xarg_t = (np.arange(2 * nmk + 1, dtype=np.int) - nmk) / sigt
#         exparg_t = (-0.5 * np.power(xarg_t, 2))
#         exparg_t = np.where(exparg_t < -20.0, -20.0, exparg_t)
#         mtkern = np.exp(exparg_t)
#         mtkern = mtkern / np.sum(mtkern)
#     else:
#         mtkern = np.zeros(2 * nmk + 1)
#         mtkern[nmk] = 1.0
#
#     mkern = area_r * np.array(mrkern) + area_t * np.array(mtkern)
#
#     return mkern, nmk



import numpy as np
from SpotDIPy import SpotDIPy
import multiprocessing
import pickle


if __name__ == '__main__':
    # multiprocessing.cpu_count() - 1
    DIP = SpotDIPy(cpu_num=multiprocessing.cpu_count() - 2, platform_name='cpu')

    DIP.set_param('t0', value=55006.4976)
    DIP.set_param('period', value= 3.4693)
    DIP.set_param('Tphot', value=5250)
    DIP.set_param('Tcool', value=4250)
    DIP.set_param('Thot', value=5250)
    DIP.set_param('incl', value=75)
    DIP.set_param('vsini', value=31.0)
    # DIP.set_param('R', value=1.668)
    DIP.set_param('vrt', value=0.0)
    DIP.set_param('mass', value=1.7)
    DIP.set_param('dOmega', value=-0.090554)
    DIP.set_param('resolution', value=0)
    DIP.set_limb_darkening_params(mh=0.0, law='linear', model='mps2', mu_min=0.1, data_path="/home/eng/Storage/ExoTiC-LD_data_v3.1.2")

    DIP.set_conf({
        'line': {'mode': 'off',
                 'wave_range': [4412, 7838],
                 'eqw': 0.0848368,
                 'scaling': {'method': 'mean'},
                 'corr': {'rv': None, 'amp': None}
                 },
        'lc': {'mode': 'on',
               'passband': 'Kepler',
               'scaling': {'method': 'mean'},
               'corr': {'amp': None}
               }
    })

    # DIP.construct_surface_grid(method='phoebe2_marching', noes=4000)
    # DIP.construct_surface_grid(method='healpy', nside=16)
    DIP.construct_surface_grid(method='trapezoid', nlats=10)

    llp_vels = np.loadtxt('teff5500_logg4.3_mh0.00_vmic1.7_lsd.out', skiprows=2)[:, 0]
    llp_phot_int = np.loadtxt('teff5500_logg4.3_mh0.00_vmic1.7_lsd.out', skiprows=2)[:, 1]
    llp_cool_int = np.loadtxt('teff4500_logg4.3_mh0.00_vmic1.7_lsd.out',  skiprows=2)[:, 1]
    llp_hot_int = np.loadtxt('teff6000_logg4.3_mh0.00_vmic1.7_lsd.out', skiprows=2)[:, 1]

    DIP.set_local_profiles({'line': {'lp_vels': llp_vels, 'phot_lp_data': llp_phot_int, 'cool_lp_data': llp_cool_int,
                                'hot_lp_data': llp_hot_int}})


    """ Make a test """
    lats_spots = [20., 40.]  # spot latitudes (degrees)
    longs_spots = [0., 100.]  # spot longitudes (degrees)
    rs_spots = [12., 10.]  # spot radius (degrees)
    cs_cools = [0.8, 0.8]  # cool spot contrast between 0 and 1
    cs_hots = [0.0, 0.0]  # hot spot contrast between 0 and 1

    spots_params = {'lats_spots': lats_spots, 'longs_spots': longs_spots, 'rs_spots': rs_spots, 'cs_cools': cs_cools,
                    'cs_hots': cs_hots}

    line_phases = np.arange(0, 1.0, 0.1)
    # line_phases = np.linspace(0, 1.0, 770)
    line_times = DIP.params['t0'] + DIP.params['period'] * line_phases
    line_vels = np.arange(-60, 60 + 1.75, 1.75)
    line_snr = 3000

    # lc_phases = np.arange(0, 1, 0.01)
    lc_phases = np.linspace(0, 360, 360*100)
    lc_times = DIP.params['t0'] + DIP.params['period'] * lc_phases
    lc_snr = 3000

    opt_params = {'alpha': 1.0, 'beta': 1.0, 'gamma': 1.0, 'delta': 1.0, 'lmbd': 1, 'maxiter': 2500,
                  'tol': 1e-5, 'disp': True}

    modes_inp = {'line': {'times': line_times, 'vels': line_vels, 'snr': line_snr},
                 'lc': {'times': lc_times, 'snr': lc_snr}}


    plot_params = {'line_sep_prf': 0.06, 'line_sep_res': 0.01, 'mol_sep_prf': 0.03, 'mol_sep_res': 0.01,
                   'show_err_bars': True, 'fmt': '%0.3f', 'markersize': 2, 'linewidth': 1, 'fontsize': 15,
                   'ticklabelsize': 12}

    # import utils as dipu
    # dipu.draw_3D_surf(DIP.params, DIP.surface_grid, spots_params)

    DIP.test(spots_params, modes_input=modes_inp, opt_params=opt_params, plot_params=plot_params)
    """"""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""

    # file = open('test_idc.pkl', 'rb')
    # input_data_dict = pickle.load(file)
    # file.close()
    #
    # DIP.set_input_data(input_data_dict)
    #
    # recons_result = DIP.reconstructor(alpha=1.0, beta=1.0, gamma=1.0, delta=1.0, lmbd=5.0, maxiter=2500, tol=1e-7,
    #                             disp=True)
    #
    # DIP.plot(plot_params={'line_sep_prf': 0.06, 'line_sep_res': 0.01, 'mol_sep_prf': 0.3, 'mol_sep_res': 0.2,
    #                       'show_err_bars': True, 'fmt': '%0.3f', 'markersize': 2, 'linewidth': 1, 'fontsize': 15,
    #                       'ticklabelsize': 12})

